#!/usr/bin/env python
# coding: utf-8
"""Tools to use xarray Datasets as inputs and outputs.

.. todo:: This module could be tidied up a bit and some methods merged or
          uniformized.
"""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging
from datetime import datetime
from itertools import product
from typing import Sequence, List, Union

import numpy as np
import pandas as pd
import xarray as xr
from scipy.optimize import newton

from capytaine import __version__
from capytaine.bodies.bodies import FloatingBody
from capytaine.bem.problems_and_results import (
    LinearPotentialFlowProblem, DiffractionProblem, RadiationProblem,
    LinearPotentialFlowResult, _default_parameters)
from capytaine.post_pro.kochin import compute_kochin
from capytaine.io.bemio import dataframe_from_bemio


LOG = logging.getLogger(__name__)


#########################
#  Reading test matrix  #
#########################

def problems_from_dataset(dataset: xr.Dataset,
                          bodies: Union[FloatingBody, Sequence[FloatingBody]],
                          ) -> List[LinearPotentialFlowProblem]:
    """Generate a list of problems from a test matrix.

    Parameters
    ----------
    dataset : xarray Dataset
        Test matrix containing the problems parameters.
    bodies : FloatingBody or list of FloatingBody
        The bodies on which the computations of the test matrix will be applied.
        They should all have different names.

    Returns
    -------
    list of LinearPotentialFlowProblem

    Raises
    ------
    ValueError
        if required fields are missing in the dataset
    """
    if isinstance(bodies, FloatingBody):
        bodies = [bodies]

    # SANITY CHECKS
    assert len(list(set(body.name for body in bodies))) == len(bodies), \
        "All bodies should have different names."

    # Warn user in case of key with unrecognized name (e.g. mispells)
    keys_in_dataset = set(dataset.keys()) | set(dataset.coords.keys())
    accepted_keys = {'wave_direction', 'radiating_dof', 'body_name', 'omega', 'water_depth', 'rho', 'g'}
    unrecognized_keys = keys_in_dataset.difference(accepted_keys)
    if len(unrecognized_keys) > 0:
        LOG.warning(f"Unrecognized key(s) in dataset: {unrecognized_keys}")

    if ("radiating_dof" not in keys_in_dataset) and ("wave_direction" not in keys_in_dataset):
        raise ValueError("Neither 'radiating_dof' nor 'wave_direction' has been provided in the dataset. "
                "No linear potential flow problem can be inferred.")
    # END SANITY CHECKS

    dataset = _unsqueeze_dimensions(dataset)

    omega_range = dataset['omega'].data if 'omega' in dataset else [_default_parameters['omega']]
    water_depth_range = dataset['water_depth'].data if 'water_depth' in dataset else [_default_parameters['water_depth']]
    rho_range = dataset['rho'].data if 'rho' in dataset else [_default_parameters['rho']]
    g_range = dataset['g'].data if 'g' in dataset else [_default_parameters['g']]

    wave_direction_range = dataset['wave_direction'].data if 'wave_direction' in dataset else None
    radiating_dofs = dataset['radiating_dof'].data.astype(object) if 'radiating_dof' in dataset else None
    # astype(object) is meant to convert Numpy internal string type numpy.str_ to Python general string type.

    if 'body_name' in dataset:
        assert set(dataset['body_name'].data) <= {body.name for body in bodies}, \
            "Some body named in the dataset was not given as argument to `problems_from_dataset`."
        body_range = {body.name: body for body in bodies if body.name in dataset['body_name'].data}
        # Only the bodies listed in the dataset have been kept
    else:
        body_range = {body.name: body for body in bodies}

    problems = []
    if wave_direction_range is not None:
        for omega, wave_direction, water_depth, body_name, rho, g \
                in product(omega_range, wave_direction_range, water_depth_range, body_range, rho_range, g_range):
            problems.append(
                DiffractionProblem(body=body_range[body_name], omega=omega,
                                   wave_direction=wave_direction, sea_bottom=-water_depth, rho=rho, g=g)
            )

    if radiating_dofs is not None:
        for omega, radiating_dof, water_depth, body_name, rho, g \
                in product(omega_range, radiating_dofs, water_depth_range, body_range, rho_range, g_range):
            problems.append(
                RadiationProblem(body=body_range[body_name], omega=omega,
                                 radiating_dof=radiating_dof, sea_bottom=-water_depth, rho=rho, g=g)
            )

    return sorted(problems)


def _squeeze_dimensions(data_array, dimensions=None):
    """Remove dimensions if they are of size 1. The coordinates become scalar coordinates."""
    if dimensions is None:
        dimensions = data_array.dims
    for dim in dimensions:
        if len(data_array[dim]) == 1:
            data_array = data_array.squeeze(dim, drop=False)
    return data_array


def _unsqueeze_dimensions(data_array, dimensions=None):
    """Add scalar coordinates as dimensions of size 1."""
    if dimensions is None:
        dimensions = list(data_array.coords.keys())
    for dim in dimensions:
        if len(data_array.coords[dim].values.shape) == 0:
            data_array = xr.concat([data_array], dim=dim)
    return data_array


######################
#  Dataset creation  #
######################

def _dataset_from_dataframe(df: pd.DataFrame,
                            variables: Union[str, Sequence[str]],
                            dimensions: Sequence[str],
                            optional_dims: Sequence[str],
                            ) -> Union[xr.DataArray, xr.Dataset]:
    """Transform a pandas.Dataframe into a xarray.Dataset.

    Parameters
    ----------
    df: pandas.DataFrame
        the input dataframe
    variables: string or sequence of strings
        the variables that will be stored in the output dataset.
        If a single name is provided, a DataArray of this variable will be provided instead.
    dimensions: sequence of strings
        Names of dimensions the variables depends on.
        They will always appear as dimension in the output dataset.
    optional_dims: sequence of strings
        Names of dimensions the variables depends on.
        They will appears as dimension in the output dataset only if they have
        more than one different values.
    """

    for variable_name in variables:
        df = df[df[variable_name].notnull()].dropna(axis='columns')  # Keep only records with non null values of all the variables
    df = df.drop_duplicates()
    df = df.set_index(optional_dims + dimensions)

    da = df.to_xarray()[variables]
    da = _squeeze_dimensions(da, dimensions=optional_dims)
    return da


def wavenumber_data_array(results: Sequence[LinearPotentialFlowResult]) -> xr.DataArray:
    """Read the wavenumbers in a list of :class:`LinearPotentialFlowResult`
    and store them into a :class:`xarray.DataArray`.
    """
    records = pd.DataFrame(
        [dict(g=result.g, water_depth=result.depth, omega=result.omega, wavenumber=result.wavenumber)
         for result in results]
    )
    ds = _dataset_from_dataframe(records, variables=['wavenumber'], dimensions=['omega'], optional_dims=['g', 'water_depth'])
    return ds['wavenumber']


def hydrostatics_dataset(bodies: Sequence[FloatingBody]) -> xr.Dataset:
    """Create a dataset by looking for 'inertia_matrix' and 'hydrostatic_stiffness'
    for each of the bodies in the list passed as argument.
    """
    dataset = xr.Dataset()
    for body_property in ['inertia_matrix', 'hydrostatic_stiffness']:
        bodies_properties = {body.name: body.__getattribute__(body_property) for body in bodies if hasattr(body, body_property)}
        if len(bodies_properties) > 0:
            bodies_properties = xr.concat(bodies_properties.values(), pd.Index(bodies_properties.keys(), name='body_name'))
            bodies_properties = _squeeze_dimensions(bodies_properties, dimensions=['body_name'])
            dataset = xr.merge([dataset, {body_property: bodies_properties}])
    return dataset


def kochin_data_array(results: Sequence[LinearPotentialFlowResult],
                      theta_range: Sequence[float],
                      **kwargs,
                      ) -> xr.Dataset:
    """Compute the Kochin function for a list of results and fills a dataset.

    .. seealso::
        :meth:`~capytaine.post_pro.kochin.compute_kochin`
            The present function is just a wrapper around :code:`compute_kochin`.
    """
    records = pd.DataFrame([
        dict(**result.problem._asdict(), theta=theta, kochin=kochin)
        for result in results
        for theta, kochin in zip(theta_range.data,
                                 compute_kochin(result, theta_range, **kwargs))
    ])

    kochin_data = {}

    if 'wave_direction' in records.columns:
        diffraction = _dataset_from_dataframe(
            records[~records['wave_direction'].isnull()],
            ['kochin'],
            dimensions=['omega', 'wave_direction', 'theta'],
            optional_dims=['g', 'rho', 'body_name', 'water_depth']
        )
        kochin_data['kochin_diffraction'] = diffraction['kochin']

    if 'radiating_dof' in records.columns:
        radiation = _dataset_from_dataframe(
            records[~records['radiating_dof'].isnull()],
            variables=['kochin'],
            dimensions=['omega', 'radiating_dof', 'theta'],
            optional_dims=['g', 'rho', 'body_name', 'water_depth']
        )
        kochin_data['kochin'] = radiation['kochin']

    return kochin_data

def collect_records(results):
    records_list = []
    warned_once_about_no_free_surface = False
    for result in results:
        if result.free_surface == np.infty:
            if not warned_once_about_no_free_surface:
                LOG.warning("Datasets currently only support cases with a free surface (free_surface=0.0).\n"
                            "Cases without a free surface (free_surface=infty) are ignored.\n"
                            "See also https://github.com/mancellin/capytaine/issues/88")
                warned_once_about_no_free_surface = True
            else:
                pass
        else:
            for record in result.records:
                records_list.append(record)
    return records_list

def assemble_dataset(results,
                     wavenumber=False, wavelength=False, mesh=False, hydrostatics=True,
                     attrs=None) -> xr.Dataset:
    """Transform a list of :class:`LinearPotentialFlowResult` into a :class:`xarray.Dataset`.

    .. todo:: The :code:`mesh` option to store information on the mesh could be improved.
              It could store the full mesh in the dataset to ensure the reproducibility of
              the results.

    Parameters
    ----------
    results: list of LinearPotentialFlowResult
        The results that will be read.
    wavenumber: bool, optional
        If True, the coordinate 'wavenumber' will be added to the output dataset.
    wavelength: bool, optional
        If True, the coordinate 'wavelength' will be added to the output dataset.
    mesh: bool, optional
        If True, store some infos on the mesh in the output dataset.
    hydrostatics: bool, optional
        If True, store the hydrostatic data in the output dataset if they exist.
    attrs: dict, optional
        Attributes that should be added to the output dataset.
    """
    dataset = xr.Dataset()

    error_msg = 'results must be either of type LinearPotentialFlowResult or a bemio.io object'
    if hasattr(results, '__iter__'):
        try:
            if 'capytaine' in results[0].__module__:
                bemio_import = False
            else:
                raise TypeError(error_msg)
        except:
            raise TypeError(error_msg)

    else:
        try:
            if 'bemio.io' in results.__module__:
                bemio_import = True
            else:
                raise TypeError(error_msg)
        except:
            raise TypeError(error_msg)
    
    if bemio_import:
        records = dataframe_from_bemio(results, wavenumber, wavelength) # TODO add hydrostatics
        all_dofs_in_order = {'Surge': None, 'Sway': None, 'Heave': None, 'Roll': None, 'Pitch': None, 'Yaw': None}

    else:
        records = pd.DataFrame(collect_records(results))
        all_dofs_in_order = {k: None for r in results for k in r.body.dofs.keys()}

    if attrs is None:
        attrs = {}
    attrs['creation_of_dataset'] = datetime.now().isoformat()
    if len(records) == 0:
        raise ValueError("No result passed to assemble_dataset.")

    inf_dof_cat = pd.CategoricalDtype(categories=all_dofs_in_order.keys())
    records["influenced_dof"] = records["influenced_dof"].astype(inf_dof_cat)
    rad_dof_cat = pd.CategoricalDtype(categories=all_dofs_in_order.keys())
    if 'added_mass' in records.columns:
        records["radiating_dof"] = records["radiating_dof"].astype(rad_dof_cat)

    optional_dims = ['g', 'rho', 'body_name', 'water_depth']

    # RADIATION RESULTS
    if 'added_mass' in records.columns:
        radiation_cases = _dataset_from_dataframe(
            records,
            variables=['added_mass', 'radiation_damping'],
            dimensions=['omega', 'radiating_dof', 'influenced_dof'],
            optional_dims=optional_dims)
        dataset = xr.merge([dataset, radiation_cases])

    # DIFFRACTION RESULTS
    if 'diffraction_force' in records.columns:
        conventions = set(records['convention'].dropna())
        if len(conventions) > 1:
            LOG.warning("Assembling a dataset mixing several conventions.")
        else:
            attrs['incoming_waves_convention'] = conventions.pop()

        diffraction_cases = _dataset_from_dataframe(
            records,
            variables=['diffraction_force', 'Froude_Krylov_force'],
            dimensions=['omega', 'wave_direction', 'influenced_dof'],
            optional_dims=optional_dims)
        dataset = xr.merge([dataset, diffraction_cases])

    # WAVENUMBER
    if wavenumber:
        if bemio_import:
            wavenumber_ds = _dataset_from_dataframe(
                records.drop_duplicates(subset=['omega']),
                variables=['wavenumber'],
                dimensions=['omega'],
                optional_dims=['g', 'water_depth'])
            dataset.coords['wavenumber'] = wavenumber_ds['wavenumber']
        else:
            dataset.coords['wavenumber'] = wavenumber_data_array(results)

    if wavelength:
        if bemio_import:
            wavelength_ds = _dataset_from_dataframe(
                    records.drop_duplicates(subset=['omega']),
                    variables=['wavelength'],
                    dimensions=['omega'],
                    optional_dims=['g', 'water_depth'])
            dataset.coords['wavelength'] = wavelength_ds['wavelength']
        else:
            dataset.coords['wavelength'] = 2*np.pi/wavenumber_data_array(results)

    if mesh:
        if bemio_import:
            LOG.warning('Bemio data does not include mesh data. mesh=True is ignored.')
        else:
            # TODO: Store full mesh...
            bodies = list({result.body for result in results})  # Filter out duplicate bodies in the list of results
            nb_faces = {body.name: body.mesh.nb_faces for body in bodies}

            def name_or_str(c):
                return c.name if hasattr(c, 'name') else str(c)
            quad_methods = {body.name: name_or_str(body.mesh.quadrature_method) for body in bodies}

            if len(nb_faces) > 1:
                dataset.coords['nb_faces'] = ('body_name', [nb_faces[name] for name in dataset.coords['body_name'].data])
                dataset.coords['quadrature_method'] = ('body_name', [quad_methods[name] for name in dataset.coords['body_name'].data])
            else:
                def the_only(d):
                    """Return the only element of a 1-element dictionary"""
                    return next(iter(d.values()))
                dataset.coords['nb_faces'] = the_only(nb_faces)
                dataset.coords['quadrature_method'] = the_only(quad_methods)

    # HYDROSTATICS
    if hydrostatics:
        if bemio_import:
            LOG.warning('Bemio data import being used, hydrostatics=True is ignored.')
        else:
            bodies = list({result.body for result in results})
            dataset = xr.merge([dataset, hydrostatics_dataset(bodies)])

    dataset.attrs.update(attrs)
    dataset.attrs['capytaine_version'] = __version__
    return dataset


################################
#  Handling of complex values  #
################################

def separate_complex_values(ds: xr.Dataset) -> xr.Dataset:
    """Return a new Dataset where complex-valued arrays of shape (...)
    have been replaced by real-valued arrays of shape (2, ...).

    .. seealso::
        :func:`merge_complex_values`
            The invert operation
    """
    ds = ds.copy()
    for variable in ds.data_vars:
        if ds[variable].dtype == complex:
            da = ds[variable]
            new_da = xr.DataArray(np.asarray((np.real(da).data, np.imag(da).data)),
                                  dims=('complex',) + da.dims)
            ds[variable] = new_da
            ds.coords['complex'] = ['re', 'im']
    return ds


def merge_complex_values(ds: xr.Dataset) -> xr.Dataset:
    """Return a new Dataset where real-valued arrays of shape (2, ...)
    have been replaced by complex-valued arrays of shape (...).

    .. seealso::
        :func:`separate_complex_values`
            The invert operation
    """
    if 'complex' in ds.coords:
        ds = ds.copy()
        for variable in ds.data_vars:
            if 'complex' in ds[variable].coords:
                da = ds[variable]
                new_dims = [d for d in da.dims if d != 'complex']
                new_da = xr.DataArray(da.sel(complex='re').data + 1j*da.sel(complex='im').data, dims=new_dims)
                ds[variable] = new_da
        ds = ds.drop_vars('complex')
    return ds



#!/usr/bin/env python
# coding: utf-8
"""Solver for the BEM problem.

Example
-------

::

    problem = RadiationProblem(...)
    result = BEMSolver(green_functions=..., engine=...).solve(problem)

"""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging

import numpy as np

from datetime import datetime

from capytaine.bem.problems_and_results import LinearPotentialFlowProblem
from capytaine.green_functions.delhommeau import Delhommeau
from capytaine.bem.engines import BasicMatrixEngine, HierarchicalToeplitzMatrixEngine
from capytaine.io.xarray import problems_from_dataset, assemble_dataset, kochin_data_array
from capytaine.tools.optional_imports import silently_import_optional_dependency

LOG = logging.getLogger(__name__)

class BEMSolver:
    """
    Solver for linear potential flow problems.

    Parameters
    ----------
    green_function: AbstractGreenFunction
        Object handling the computation of the Green function.
    engine: MatrixEngine
        Object handling the building of matrices and the resolution of linear systems with these matrices.

    Attributes
    ----------
    exportable_settings : dict
        Settings of the solver that can be saved to reinit the same solver later.
    """

    def __init__(self, *, green_function=Delhommeau(), engine=BasicMatrixEngine()):
        self.green_function = green_function
        self.engine = engine

        try:
            self.exportable_settings = {
                **self.green_function.exportable_settings,
                **self.engine.exportable_settings
            }
        except AttributeError:
            pass

    @classmethod
    def from_exported_settings(settings):
        raise NotImplementedError

    def solve(self, problem, keep_details=True):
        """Solve the linear potential flow problem.

        Parameters
        ----------
        problem: LinearPotentialFlowProblem
            the problem to be solved
        keep_details: bool, optional
            if True, store the sources and the potential on the floating body in the output object
            (default: True)

        Returns
        -------
        LinearPotentialFlowResult
            an object storing the problem data and its results
        """
        LOG.info("Solve %s.", problem)

        S, K = self.engine.build_matrices(
            problem.body.mesh, problem.body.mesh,
            problem.free_surface, problem.sea_bottom, problem.wavenumber,
            self.green_function
        )
        sources = self.engine.linear_solver(K, problem.boundary_condition)
        potential = S @ sources

        result = problem.make_results_container()
        if keep_details:
            result.sources = sources
            result.potential = potential

        for influenced_dof_name, influenced_dof_vectors in problem.influenced_dofs.items():
            # Scalar product on each face:
            influenced_dof_normal = np.sum(influenced_dof_vectors * problem.body.mesh.faces_normals, axis=1)
            # Sum over all faces:
            integrated_potential = - problem.rho * np.sum(potential * influenced_dof_normal * problem.body.mesh.faces_areas)
            # Store result:
            result.store_force(influenced_dof_name, integrated_potential)
            # Depending of the type of problem, the force will be kept as a complex-valued Froude-Krylov force
            # or stored as a couple of added mass and radiation damping coefficients.

        LOG.debug("Done!")

        return result

    def solve_all(self, problems, *, n_jobs=1, **kwargs):
        """Solve several problems.
        Optional keyword arguments are passed to `BEMSolver.solve`.

        Parameters
        ----------
        problems: list of LinearPotentialFlowProblem
            several problems to be solved
        n_jobs: int, optional (default: 1)
            the number of jobs to run in parallel using the optional dependency `joblib`
            By defaults: do not use joblib and solve sequentially.

        Returns
        -------
        list of LinearPotentialFlowResult
            the solved problems
        """
        if n_jobs == 1:  # force sequential resolution
            return [self.solve(pb, **kwargs) for pb in sorted(problems)]
        else:
            joblib = silently_import_optional_dependency("joblib")
            if joblib is None:
                raise ImportError(f"Setting the `n_jobs` argument to {n_jobs} requires the missing optional dependency 'joblib'.")
            groups_of_problems = LinearPotentialFlowProblem._group_for_parallel_resolution(problems)
            groups_of_results = joblib.Parallel(n_jobs=n_jobs)(joblib.delayed(self.solve_all)(grp, n_jobs=1, **kwargs) for grp in groups_of_problems)
            results = [res for grp in groups_of_results for res in grp]  # flatten the nested list
            return results

    def fill_dataset(self, dataset, bodies, *, n_jobs=1, **kwargs):
        """Solve a set of problems defined by the coordinates of an xarray dataset.

        Parameters
        ----------
        dataset : xarray Dataset
            dataset containing the problems parameters: frequency, radiating_dof, water_depth, ...
        bodies : FloatingBody or list of FloatingBody
            The body or bodies involved in the problems
            They should all have different names.
        n_jobs: int, optional (default: 1)
            the number of jobs to run in parallel using the optional dependency `joblib`
            By defaults: do not use joblib and solve sequentially.

        Returns
        -------
        xarray Dataset
        """
        attrs = {'start_of_computation': datetime.now().isoformat(),
                 **self.exportable_settings}
        problems = problems_from_dataset(dataset, bodies)
        if 'theta' in dataset.coords:
            results = self.solve_all(problems, keep_details=True, n_jobs=n_jobs)
            kochin = kochin_data_array(results, dataset.coords['theta'])
            dataset = assemble_dataset(results, attrs=attrs, **kwargs)
            dataset.update(kochin)
        else:
            results = self.solve_all(problems, keep_details=False, n_jobs=n_jobs)
            dataset = assemble_dataset(results, attrs=attrs, **kwargs)
        return dataset

    def get_potential_on_mesh(self, result, mesh, chunk_size=50):
        """Compute the potential on a mesh for the potential field of a previously solved problem.
        Since the interaction matrix does not need to be computed in full to compute the matrix-vector product,
        only a few lines are evaluated at a time to reduce the memory cost of the operation.

        Parameters
        ----------
        result : LinearPotentialFlowResult
            the return of the BEM solver
        mesh : Mesh or CollectionOfMeshes
            a mesh
        chunk_size: int, optional
            Number of lines to compute in the matrix.
            (legacy, should be passed as an engine setting instead).

        Returns
        -------
        array of shape (mesh.nb_faces,)
            potential on the faces of the mesh

        Raises
        ------
        Exception: if the :code:`Result` object given as input does not contain the source distribution.
        """
        LOG.info(f"Compute potential on {mesh.name} for {result}.")

        if result.sources is None:
            raise Exception(f"""The values of the sources of {result} cannot been found.
            They probably have not been stored by the solver because the option keep_details=True have not been set.
            Please re-run the resolution with this option.""")

        if chunk_size > mesh.nb_faces:
            S = self.engine.build_S_matrix(
                mesh,
                result.body.mesh,
                result.free_surface, result.sea_bottom, result.wavenumber,
                self.green_function
            )
            phi = S @ result.sources

        else:
            phi = np.empty((mesh.nb_faces,), dtype=np.complex128)
            for i in range(0, mesh.nb_faces, chunk_size):
                faces_to_extract = list(range(i, min(i+chunk_size, mesh.nb_faces)))
                S = self.engine.build_S_matrix(
                    mesh.extract_faces(faces_to_extract),
                    result.body.mesh,
                    result.free_surface, result.sea_bottom, result.wavenumber,
                    self.green_function
                )
                phi[i:i+chunk_size] = S @ result.sources

        LOG.debug(f"Done computing potential on {mesh.name} for {result}.")

        return phi

    def get_free_surface_elevation(self, result, free_surface, keep_details=False):
        """Compute the elevation of the free surface on a mesh for a previously solved problem.

        Parameters
        ----------
        result : LinearPotentialFlowResult
            the return of the solver
        free_surface : FreeSurface
            a meshed free surface
        keep_details : bool, optional
            if True, keep the free surface elevation in the LinearPotentialFlowResult (default:False)

        Returns
        -------
        array of shape (free_surface.nb_faces,)
            the free surface elevation on each faces of the meshed free surface

        Raises
        ------
        Exception: if the :code:`Result` object given as input does not contain the source distribution.
        """
        fs_elevation = 1j*result.omega/result.g * self.get_potential_on_mesh(result, free_surface.mesh)
        if keep_details:
            result.fs_elevation[free_surface] = fs_elevation
        return fs_elevation


# LEGACY INTERFACE

def _arguments(f):
    """Returns the name of the arguments of the function f"""
    return f.__code__.co_varnames[:f.__code__.co_argcount]

class Nemoh(BEMSolver):
    """Solver for the BEM problem based on Nemoh's Green function. Legacy API.
    Parameters are dispatched to the Delhommeau class and to the engine
    (BasicMatrixEngine or HierarchicalToeplitzMatrixEngine).
    """

    def __init__(self, **params):
        green_function = Delhommeau(
           **{key: params[key] for key in params if key in _arguments(Delhommeau.__init__)}
	)
        if 'hierarchical_matrices' in params and params['hierarchical_matrices']:
            engine = HierarchicalToeplitzMatrixEngine(
               **{key: params[key] for key in params if key in _arguments(HierarchicalToeplitzMatrixEngine.__init__)}
            )
        else:
            engine = BasicMatrixEngine(
               **{key: params[key] for key in params if key in _arguments(BasicMatrixEngine.__init__)}
            )

        super().__init__(green_function=green_function, engine=engine)

    def build_matrices(self, *args, **kwargs):
        """Legacy API."""
        args = args + (self.green_function,)
        return self.engine.build_matrices(*args, **kwargs)




#!/usr/bin/env python
# coding: utf-8
"""Floating bodies to be used in radiation-diffraction problems."""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging
import copy
from itertools import chain, accumulate, product, zip_longest

import datetime

import numpy as np
import xarray as xr

from capytaine.meshes.geometry import Abstract3DObject, Plane, inplace_transformation
from capytaine.meshes.meshes import Mesh
from capytaine.meshes.symmetric import build_regular_array_of_meshes
from capytaine.meshes.collections import CollectionOfMeshes

LOG = logging.getLogger(__name__)

TRANSLATION_DOFS_DIRECTIONS = {"surge": (1, 0, 0), "sway": (0, 1, 0), "heave": (0, 0, 1)}
ROTATION_DOFS_AXIS = {"roll": (1, 0, 0), "pitch": (0, 1, 0), "yaw": (0, 0, 1)}


class FloatingBody(Abstract3DObject):
    """A floating body described as a mesh and some degrees of freedom.

    The mesh structure is stored as a Mesh from capytaine.mesh.mesh or a
    CollectionOfMeshes from capytaine.mesh.meshes_collection.

    The degrees of freedom (dofs) are stored as a dict associating a name to
    a complex-valued array of shape (nb_faces, 3). To each face of the body
    (as indexed in the mesh) corresponds a complex-valued 3d vector, which
    defines the displacement of the center of the face in frequency domain.

    Parameters
    ----------
    mesh : Mesh or CollectionOfMeshes, optional
        the mesh describing the geometry of the floating body.
        If none is given, a empty one is created.
    dofs : dict, optional
        the degrees of freedom of the body.
        If none is given, a empty dictionary is initialized.
    mass : float or None, optional
        the mass of the body in kilograms.
        Required only for some hydrostatics computation.
        If None, the mass is implicitly assumed to be the mass of displaced water.
    center_of_mass: 3-element array, optional
        the position of the center of mass.
        Required only for some hydrostatics computation.
    name : str, optional
        a name for the body.
        If none is given, the one of the mesh is used.
    """

    def __init__(self, mesh=None, dofs=None, mass=None, center_of_mass=None, name=None):
        if mesh is None:
            mesh = Mesh(name="dummy_mesh")

        if dofs is None:
            dofs = {}

        if name is None:
            name = mesh.name

        assert isinstance(mesh, Mesh) or isinstance(mesh, CollectionOfMeshes)
        self.mesh = mesh
        self.full_body = None
        self.dofs = dofs
        self.mass = mass
        self.center_of_mass = center_of_mass
        self.name = name

        if self.mesh.nb_vertices == 0 or self.mesh.nb_faces == 0:
            LOG.warning(f"New floating body (with empty mesh!): {self.name}.")
        else:
            self.mesh.heal_mesh()
            LOG.info(f"New floating body: {self.name}.")

    @staticmethod
    def from_meshio(mesh, name=None) -> 'FloatingBody':
        """Create a FloatingBody from a meshio mesh object."""

        import meshio
        if not isinstance(mesh, meshio._mesh.Mesh):
            raise TypeError('mesh must be of type meshio._mesh.Mesh, received {:}'.format(type(mesh)))

        if name is None:
            date_str = datetime.datetime.now().strftime('%Y%m%d%H%M%S%f')
            name = 'fb_{:}'.format(date_str)

        def all_faces_as_quads(cells):
            all_faces = []
            if 'quad' in cells:
                all_faces.append(cells['quad'])
            if 'triangle' in cells:
                num_triangles = len(mesh.cells_dict['triangle'])
                LOG.info("Stored {:} triangle faces as quadrilaterals".format(num_triangles))
                triangles_as_quads = np.empty((cells['triangle'].shape[0], 4), dtype=int)
                triangles_as_quads[:, :3] = cells['triangle'][:, :]
                triangles_as_quads[:, 3] = cells['triangle'][:, 2]  # Repeat one node to make a quad
                all_faces.append(triangles_as_quads)
            return np.concatenate(all_faces)

        cpt_mesh = Mesh(vertices=mesh.points,
                        faces=all_faces_as_quads(mesh.cells_dict),
                        name=name+"_mesh")

        fb = FloatingBody(mesh=cpt_mesh, name=name)
        return fb


    @staticmethod
    def from_file(filename: str, file_format=None, name=None) -> 'FloatingBody':
        """Create a FloatingBody from a mesh file using meshmagick."""
        from capytaine.io.mesh_loaders import load_mesh
        if name is None:
            name = filename
        mesh = load_mesh(filename, file_format, name=f"{name}_mesh")
        return FloatingBody(mesh, name=name)

    def __lt__(self, other: 'FloatingBody') -> bool:
        """Arbitrary order. The point is to sort together the problems involving the same body."""
        return self.name < other.name

    ##########
    #  Dofs  #
    ##########

    @property
    def nb_dofs(self) -> int:
        """Number of degrees of freedom."""
        return len(self.dofs)

    def add_translation_dof(self, direction=None, name=None, amplitude=1.0) -> None:
        """Add a new translation dof (in place).
        If no direction is given, the code tries to infer it from the name.

        Parameters
        ----------
        direction : array of shape (3,), optional
            the direction of the translation
        name : str, optional
            a name for the degree of freedom
        amplitude : float, optional
            amplitude of the dof (default: 1.0 m/s)
        """
        if direction is None:
            if name is not None and name.lower() in TRANSLATION_DOFS_DIRECTIONS:
                direction = TRANSLATION_DOFS_DIRECTIONS[name.lower()]
            else:
                raise ValueError("A direction needs to be specified for the dof.")

        if name is None:
            name = f"dof_{self.nb_dofs}_translation"

        direction = np.asarray(direction)
        assert direction.shape == (3,)

        motion = np.empty((self.mesh.nb_faces, 3))
        motion[:, :] = direction
        self.dofs[name] = amplitude * motion

    def add_rotation_dof(self, axis=None, name=None, amplitude=1.0) -> None:
        """Add a new rotation dof (in place).
        If no axis is given, the code tries to infer it from the name.

        Parameters
        ----------
        axis: Axis, optional
            the axis of the rotation
        name : str, optional
            a name for the degree of freedom
        amplitude : float, optional
            amplitude of the dof (default: 1.0)
        """
        if axis is None:
            if name is not None and name.lower() in ROTATION_DOFS_AXIS:
                axis_direction = ROTATION_DOFS_AXIS[name.lower()]
                for point_attr in ('rotation_center', 'center_of_mass', 'geometric_center'):
                    if hasattr(self, point_attr) and getattr(self, point_attr) is not None:
                        axis_point = getattr(self, point_attr)
                        LOG.info(f"The rotation dof {name} has been initialized around the point: "
                                 f"{self.name}.{point_attr} = {getattr(self, point_attr)}")
                        break
                else:
                    axis_point = np.array([0, 0, 0])
                    LOG.warning(f"The rotation dof {name} has been initialized "
                                f"around the origin of the domain (0, 0, 0).")
            else:
                raise ValueError("A direction needs to be specified for the dof.")
        else:
            axis_point = axis.point
            axis_direction = axis.vector

        if name is None:
            name = f"dof_{self.nb_dofs}_rotation"

        if self.mesh.nb_faces == 0:
            self.dofs[name] = np.empty((self.mesh.nb_faces, 3))
        else:
            motion = np.cross(axis_point - self.mesh.faces_centers, axis_direction)
            self.dofs[name] = amplitude * motion

    def add_all_rigid_body_dofs(self) -> None:
        """Add the six degrees of freedom of rigid bodies (in place)."""
        self.add_translation_dof(name="Surge")
        self.add_translation_dof(name="Sway")
        self.add_translation_dof(name="Heave")
        self.add_rotation_dof(name="Roll")
        self.add_rotation_dof(name="Pitch")
        self.add_rotation_dof(name="Yaw")

    @inplace_transformation
    def keep_only_dofs(self, dofs):
        for dof in list(self.dofs.keys()):
            if dof not in dofs:
                del self.dofs[dof]

        if hasattr(self, 'inertia_matrix'):
            self.inertia_matrix = self.inertia_matrix.sel(radiating_dof=dofs, influenced_dof=dofs)
        if hasattr(self, 'hydrostatic_stiffness'):
            self.hydrostatic_stiffness = self.hydrostatic_stiffness.sel(radiating_dof=dofs, influenced_dof=dofs)

        return self

    def add_dofs_labels_to_vector(self, vector):
        """Helper function turning a bare vector into a vector labelled by the name of the dofs of the body,
        to be used for instance for the computation of RAO."""
        return xr.DataArray(data=np.asarray(vector), dims=['influenced_dof'],
                            coords={'influenced_dof': list(self.dofs)},
                            )

    def add_dofs_labels_to_matrix(self, matrix):
        """Helper function turning a bare matrix into a matrix labelled by the name of the dofs of the body,
        to be used for instance for the computation of RAO."""
        return xr.DataArray(data=np.asarray(matrix), dims=['influenced_dof', 'radiating_dof'],
                            coords={'influenced_dof': list(self.dofs), 'radiating_dof': list(self.dofs)},
                            )

    ###################
    # Hydrostatics #
    ###################

    def surface_integral(self, data, **kwargs):
        """Returns integral of given data along wet surface area."""
        return np.sum(data * self.mesh.faces_areas, **kwargs)

    def waterplane_integral(self, data, **kwargs):
        """Returns integral of given data along water plane area."""
        return self.surface_integral(self.mesh.faces_normals[:,2] * data, **kwargs)

    @property
    def wet_surface_area(self):
        """Returns wet surface area."""
        return self.surface_integral(1)

    @property
    def volumes(self):
        """Returns volumes using x, y, z components of the FloatingBody."""
        norm_coord = self.mesh.faces_normals * self.mesh.faces_centers
        return self.surface_integral(norm_coord.T, axis=1)

    @property
    def volume(self):
        """Returns volume of the FloatingBody."""
        return np.mean(self.volumes)

    def disp_mass(self, *, rho=1000):
        return rho * self.volume

    @property
    def center_of_buoyancy(self):
        """Returns center of buoyancy of the FloatingBody."""
        volume = self.volume
        coords_sq_norm = self.mesh.faces_normals * self.mesh.faces_centers**2
        return self.surface_integral(coords_sq_norm.T, axis=1) / (2*volume)

    @property
    def waterplane_area(self):
        """Returns water plane area of the FloatingBody."""
        waterplane_area = -self.waterplane_integral(1)
        return waterplane_area

    @property
    def waterplane_center(self):
        """Returns water plane center of the FloatingBody.

        Note: Returns None if the FloatingBody is full submerged.
        """
        waterplane_area = self.waterplane_area
        if abs(waterplane_area) < 1e-10:
            return None
        else:
            waterplane_center = -self.waterplane_integral(
                self.mesh.faces_centers.T, axis=1) / waterplane_area
            return waterplane_center[:-1]

    @property
    def transversal_metacentric_radius(self):
        """Returns transversal metacentric radius of the body."""
        inertia_moment = -self.waterplane_integral(self.mesh.faces_centers[:,1]**2)
        return inertia_moment / self.volume

    @property
    def longitudinal_metacentric_radius(self):
        """Returns longitudinal metacentric radius of the body."""
        inertia_moment = -self.waterplane_integral(self.mesh.faces_centers[:,0]**2)
        return inertia_moment / self.volume

    @property
    def transversal_metacentric_height(self):
        """Returns transversal metacentric height of the body."""
        gb = self.center_of_mass - self.center_of_buoyancy
        return self.transversal_metacentric_radius - gb[2]

    @property
    def longitudinal_metacentric_height(self):
        """Returns longitudinal metacentric height of the body."""
        gb = self.center_of_mass - self.center_of_buoyancy
        return self.longitudinal_metacentric_radius - gb[2]

    def dof_normals(self, dof):
        """Returns dot product of the surface face normals and DOF"""
        return np.sum(self.mesh.faces_normals * dof, axis=1)

    def _infer_rotation_center(self):
        """Hacky way to infer the point around which the rotation dofs are defined.
        (Assuming all three rotation dofs are defined around the same point).
        In the future, should be replaced by something more robust.
        """
        if hasattr(self, "rotation_center"):
            return np.asarray(self.rotation_center)

        else:
            try:
                xc1 = self.dofs["Pitch"][:, 2] + self.mesh.faces_centers[:, 0]
                xc2 = -self.dofs["Yaw"][:, 1] + self.mesh.faces_centers[:, 0]
                yc1 = self.dofs["Yaw"][:, 0] + self.mesh.faces_centers[:, 1]
                yc2 = -self.dofs["Roll"][:, 2] + self.mesh.faces_centers[:, 1]
                zc1 = -self.dofs["Pitch"][:, 0] + self.mesh.faces_centers[:, 2]
                zc2 = self.dofs["Roll"][:, 1] + self.mesh.faces_centers[:, 2]

                # All items should be identical in a given vector
                assert np.isclose(xc1, xc1[0]).all()
                assert np.isclose(yc1, yc1[0]).all()
                assert np.isclose(zc1, zc1[0]).all()

                # Both vector should be identical
                assert np.allclose(xc1, xc2)
                assert np.allclose(yc1, yc2)
                assert np.allclose(zc1, zc2)

                return np.array([xc1[0], yc1[0], zc1[0]])

            except Exception as e:
                raise ValueError(
                        f"Failed to infer the rotation center of {self.name} to compute rigid body hydrostatics.\n"
                        f"Possible fix: add a `rotation_center` attibute to {self.name}.\n"
                        "Note that rigid body hydrostatic methods currently assume that the three rotation dofs have the same rotation center."
                        ) from e

    def each_hydrostatic_stiffness(self, influenced_dof_name, radiating_dof_name, *,
                                         influenced_dof_div=0.0, rho=1000.0, g=9.81):
        r"""
        Return the hydrostatic stiffness for a pair of DOFs.

        :math:`C_{ij} = \rho g\iint_S (\hat{n} \cdot V_j) (w_i + z D_i) dS`

        where :math:`\hat{n}` is surface normal,

        :math:`V_i = u_i \hat{n}_x + v_i \hat{n}_y + w_i \hat{n}_z` is DOF vector and

        :math:`D_i = \nabla \cdot V_i` is the divergence of the DOF.

        Parameters
        ----------
        influenced_dof_name : str
            Name of influenced DOF vector of the FloatingBody
        radiating_dof_name: str
            Name of radiating DOF vector of the FloatingBody
        influenced_dof_div: np.ndarray (Face_count), optional
            Influenced DOF divergence of the FloatingBody, by default 0.0.
        rho: float, optional
            water density, by default 1000.0
        g: float, optional
            Gravity acceleration, by default 9.81

        Returns
        -------
        hs_ij: xarray.variable
            hydrostatic_stiffness of ith DOF and jth DOF.

        Note
        ----
            This function computes the hydrostatic stiffness assuming :math:`D_{i} = 0`.
            If :math:`D_i \neq 0`, input the divergence interpolated to face centers.

            General integral equations are used for the rigid body modes and
            Neumann (1994) method is used for flexible modes.

        References
        ----------
            Newman, John Nicholas. "Wave effects on deformable bodies."Applied ocean
            research" 16.1 (1994): 47-59.
            http://resolver.tudelft.nl/uuid:0adff84c-43c7-43aa-8cd8-d4c44240bed8

        """
        # Newmann (1994) formula is not 'complete' as recovering the rigid body
        # terms is not possible. https://doi.org/10.1115/1.3058702.

        # Alternative is to use the general equation of hydrostatic and
        # restoring coefficient for rigid mdoes and use Neuman equation for elastic
        # modes.

        rigid_dof_names = ("Surge", "Sway", "Heave", "Roll", "Pitch", "Yaw")
        dof_pair = (influenced_dof_name, radiating_dof_name)

        if set(dof_pair).issubset(set(rigid_dof_names)):
            if self.center_of_mass is None:
                raise ValueError(f"Trying to compute rigid-body hydrostatic stiffness for {self.name}, but no center of mass has been defined.\n"
                                 f"Suggested solution: define a `center_of_mass` attribute for the FloatingBody {self.name}.")
            mass = self.disp_mass(rho=rho) if self.mass is None else self.mass
            xc, yc, zc = self._infer_rotation_center()

            if dof_pair == ("Heave", "Heave"):
                norm_hs_stiff = self.waterplane_area
            elif dof_pair in [("Heave", "Roll"), ("Roll", "Heave")]:
                norm_hs_stiff = -self.waterplane_integral(self.mesh.faces_centers[:,1] - yc)
            elif dof_pair in [("Heave", "Pitch"), ("Pitch", "Heave")]:
                norm_hs_stiff = self.waterplane_integral(self.mesh.faces_centers[:,0] - xc)
            elif dof_pair == ("Roll", "Roll"):
                norm_hs_stiff = (
                        -self.waterplane_integral((self.mesh.faces_centers[:,1] - yc)**2)
                        + self.volume*(self.center_of_buoyancy[2] - zc) - mass/rho*(self.center_of_mass[2] - zc)
                )
            elif dof_pair in [("Roll", "Pitch"), ("Pitch", "Roll")]:
                norm_hs_stiff = self.waterplane_integral((self.mesh.faces_centers[:,0] - xc)
                                                          * (self.mesh.faces_centers[:,1] - yc))
            elif dof_pair == ("Roll", "Yaw"):
                norm_hs_stiff = - self.volume*(self.center_of_buoyancy[0] - xc) + mass/rho*(self.center_of_mass[0] - xc)
            elif dof_pair == ("Pitch", "Pitch"):
                norm_hs_stiff = (
                        -self.waterplane_integral((self.mesh.faces_centers[:,0] - xc)**2)
                        + self.volume*(self.center_of_buoyancy[2] - zc) - mass/rho*(self.center_of_mass[2] - zc)
                        )
            elif dof_pair == ("Pitch", "Yaw"):
                norm_hs_stiff = - self.volume*(self.center_of_buoyancy[1] - yc) + mass/rho*(self.center_of_mass[1] - yc)
            else:
                norm_hs_stiff = 0.0
        else:
            if self.mass is not None and np.isclose(self.mass, self.disp_mass(rho), rtol=1e-4):
                raise NotImplementedError(
                        f"Trying to compute the hydrostatic stiffness for dofs {radiating_dof_name} and {influenced_dof_name}"
                        f"of body {self.name}, which is not neutrally buoyant (mass={body.mass}, disp_mass={body.disp_mass(rho)}.\n"
                        f"This case has not been implemented in Capytaine. You need either a single rigid body or a neutrally buoyant body."
                        )

            # Neuman (1994) formula for flexible DOFs
            influenced_dof = np.array(self.dofs[influenced_dof_name])
            radiating_dof = np.array(self.dofs[radiating_dof_name])
            influenced_dof_div_array = np.array(influenced_dof_div)

            radiating_dof_normal = self.dof_normals(radiating_dof)
            z_influenced_dof_div = influenced_dof[:,2] + self.mesh.faces_centers[:,2] * influenced_dof_div_array
            norm_hs_stiff = self.surface_integral( -radiating_dof_normal * z_influenced_dof_div)

        hs_stiff = rho * g * norm_hs_stiff

        return xr.DataArray([[hs_stiff]],
                            dims=['influenced_dof', 'radiating_dof'],
                            coords={'influenced_dof': [influenced_dof_name],
                            'radiating_dof': [radiating_dof_name]},
                            name="hydrostatic_stiffness"
                            )

    def compute_hydrostatic_stiffness(self, *, divergence=None, rho=1000.0, g=9.81):
        r"""
        Compute hydrostatic stiffness matrix for all DOFs of the body.

        :math:`C_{ij} = \rho g\iint_S (\hat{n} \cdot V_j) (w_i + z D_i) dS`

        where :math:`\hat{n}` is surface normal,

        :math:`V_i = u_i \hat{n}_x + v_i \hat{n}_y + w_i \hat{n}_z` is DOF vector and

        :math:`D_i = \nabla \cdot V_i` is the divergence of the DOF.

        Parameters
        ----------
        divergence : dict mapping a dof name to an array of shape (nb_faces) or
                        xarray.DataArray of shape (nb_dofs × nb_faces), optional
            Divergence of the DOFs, by default None
        rho : float, optional
            Water density, by default 1000.0
        g: float, optional
            Gravity acceleration, by default 9.81

        Returns
        -------
        xr.DataArray
            Matrix of hydrostatic stiffness

        Note
        ----
            This function computes the hydrostatic stiffness assuming :math:`D_{i} = 0`.
            If :math:`D_i \neq 0`, input the divergence interpolated to face centers.

            General integral equations are used for the rigid body modes and
            Neumann (1994) method is used for flexible modes.

        References
        ----------
            Newman, John Nicholas. "Wave effects on deformable bodies."Applied ocean
            research" 16.1 (1994): 47-59.
            http://resolver.tudelft.nl/uuid:0adff84c-43c7-43aa-8cd8-d4c44240bed8

        """
        if len(self.dofs) == 0:
            raise AttributeError("Cannot compute hydrostatics stiffness on {} since no dof has been defined.".format(self.name))

        def divergence_dof(influenced_dof):
            if divergence is None:
                return 0.0
            elif isinstance(divergence, dict) and influenced_dof in divergence.keys():
                return divergence[influenced_dof]
            elif isinstance(divergence, xr.DataArray) and influenced_dof in divergence.coords["influenced_dof"]:
                return divergence.sel(influenced_dof=influenced_dof).values
            else:
                LOG.warning("Computing hydrostatic stiffness without the divergence of {}".format(influenced_dof))
                return 0.0

        hs_set =  xr.merge([
            self.each_hydrostatic_stiffness(
                influenced_dof_name, radiating_dof_name,
                influenced_dof_div = divergence_dof(influenced_dof_name),
                rho=rho, g=g
                )
            for radiating_dof_name in self.dofs
            for influenced_dof_name in self.dofs
            ])

        # Reorder dofs
        K = hs_set.hydrostatic_stiffness.sel(influenced_dof=list(self.dofs.keys()), radiating_dof=list(self.dofs.keys()))
        return K

    def compute_rigid_body_inertia(self, *, rho=1000, output_type="body_dofs"):
        """
        Inertia Mass matrix of the body for 6 rigid DOFs.

        Parameters
        ----------
        rho : float, optional
            Density of water, by default 1000.0
        output_type : {"body_dofs", "rigid_dofs", "all_dofs"}
            Type of DOFs for mass mat output, by default "body_dofs".

        Returns
        -------
        xarray.DataArray
            Inertia matrix

        Raises
        ------
        ValueError
            If output_type is not in {"body_dofs", "rigid_dofs", "all_dofs"}.
        """
        if self.center_of_mass is None:
            raise ValueError(f"Trying to compute rigid-body inertia matrix for {self.name}, but no center of mass has been defined.\n"
                             f"Suggested solution: define a `center_of_mass` attribute for the FloatingBody {self.name}.")

        rc = self._infer_rotation_center()
        fcs = (self.mesh.faces_centers - rc).T
        combinations = np.array([fcs[0]**2, fcs[1]**2, fcs[2]**2, fcs[0]*fcs[1],
                                 fcs[1]*fcs[2], fcs[2]*fcs[0]])
        integrals = np.array([
            [np.sum(normal_i * fcs[axis] * combination * self.mesh.faces_areas)
            for combination in combinations]
            for axis, normal_i in enumerate(self.mesh.faces_normals.T)])


        inertias = np.array([
            (integrals[0,1]   + integrals[0,2]   + integrals[1,1]/3
             + integrals[1,2]   + integrals[2,1] + integrals[2,2]/3)/3,
            (integrals[0,0]/3 + integrals[0,2]   + integrals[1,0]
             + integrals[1,2]   + integrals[2,0] + integrals[2,2]/3)/3,
            (integrals[0,0]/3 + integrals[0,1]   + integrals[1,0]
             + integrals[1,1]/3 + integrals[2,0] + integrals[2,1]  )/3,
            integrals[2,3],
            integrals[0,4],
            integrals[1,5]
        ])

        cog = self.center_of_mass - rc
        volume = self.volume
        volumic_inertia_matrix = np.array([
            [ volume        , 0              , 0               ,
              0             , volume*cog[2]  , -volume*cog[1]  ],
            [ 0             , volume         , 0               ,
             -volume*cog[2] , 0              , volume*cog[0]   ],
            [ 0             , 0              , volume          ,
              volume*cog[1] , -volume*cog[0] , 0 ]             ,
            [ 0             , -volume*cog[2] , volume*cog[1]   ,
              inertias[0]   , -inertias[3]   , -inertias[5]    ],
            [ volume*cog[2] , 0              , -volume*cog[0]  ,
             -inertias[3]   , inertias[1]    , -inertias[4]    ],
            [-volume*cog[1] , volume*cog[0]  , 0               ,
             -inertias[5]   , -inertias[4]   , inertias[2]     ],
        ])

        density = rho if self.mass is None else self.mass/volume
        inertia_matrix = density * volumic_inertia_matrix

        # Rigid DOFs
        rigid_dof_names = ["Surge", "Sway", "Heave", "Roll", "Pitch", "Yaw"]
        rigid_inertia_matrix_xr = xr.DataArray(data=np.asarray(inertia_matrix),
                            dims=['influenced_dof', 'radiating_dof'],
                            coords={'influenced_dof': rigid_dof_names,
                                    'radiating_dof': rigid_dof_names},
                            name="inertia_matrix")

        # Body DOFs (Default as np.nan)
        body_dof_names = list(self.dofs)
        body_dof_count = len(body_dof_names)
        other_dofs_inertia_matrix_xr = xr.DataArray(np.nan * np.zeros([body_dof_count, body_dof_count]),
                                    dims=['influenced_dof', 'radiating_dof'],
                                    coords={'influenced_dof': body_dof_names,
                                            'radiating_dof': body_dof_names},
                                    name="inertia_matrix")

        total_mass_xr = xr.merge([rigid_inertia_matrix_xr, other_dofs_inertia_matrix_xr], compat="override").inertia_matrix

        non_rigid_dofs = set(body_dof_names) - set(rigid_dof_names)

        if output_type == "body_dofs":
            if len(non_rigid_dofs) > 0:
                LOG.warning(f"Non-rigid dofs: {non_rigid_dofs} are detected and \
respective inertia coefficients are assigned as NaN.")

            inertia_matrix_xr = total_mass_xr.sel(influenced_dof=body_dof_names,
                                                  radiating_dof=body_dof_names)
        elif output_type == "rigid_dofs":
            inertia_matrix_xr = total_mass_xr.sel(influenced_dof=rigid_dof_names,
                                                  radiating_dof=rigid_dof_names)
        elif output_type == "all_dofs":
            if len(non_rigid_dofs) > 0:
                LOG.warning("Non-rigid dofs: {non_rigid_dofs} are detected and \
respective inertia coefficients are assigned as NaN.")

            inertia_matrix_xr = total_mass_xr
        else:
            raise ValueError(f"output_type should be either 'body_dofs', \
'all_dofs' or 'rigid_dofs'. Given output_type = '{output_type}'.")

        return inertia_matrix_xr


    def compute_hydrostatics(self, *, rho=1000.0, g=9.81, divergence=None):
        """Compute hydrostatics of the FloatingBody.

        Parameters
        ----------
        rho : float, optional
            Density of Water. The default is 1000.
        g: float, optional
            Gravity acceleration. The default is 9.81.
        divergence : np.ndarray, optional
            Divergence of the DOFs.

        Returns
        -------
        hydrostatics : dict
            All hydrostatics values of the FloatingBody.
        """
        if self.center_of_mass is None:
            raise ValueError(f"Trying to compute hydrostatics for {self.name}, but no center of mass has been defined.\n"
                             f"Suggested solution: define a `center_of_mass` attribute for the FloatingBody {self.name}.")

        self.keep_immersed_part()

        full_mesh_vertices = self.full_body.mesh.vertices
        coord_max = full_mesh_vertices.max(axis=0)
        coord_min = full_mesh_vertices.min(axis=0)
        full_length, full_breadth, depth = full_mesh_vertices.max(axis=0) - full_mesh_vertices.min(axis=0)

        vertices = self.mesh.vertices
        sub_length, sub_breadth, _ = vertices.max(axis=0) - vertices.min(axis=0)

        if abs(self.waterplane_area) > 1e-10:
            water_plane_idx = np.isclose(vertices[:,2], 0.0)
            water_plane = vertices[water_plane_idx][:,:-1]
            wl_length, wl_breadth = water_plane.max(axis=0) - water_plane.min(axis=0)
        else:
            wl_length, wl_breadth = 0.0, 0.0

        hydrostatics = {}
        hydrostatics["g"] = g
        hydrostatics["rho"] = rho
        hydrostatics["center_of_mass"] = self.center_of_mass

        hydrostatics["wet_surface_area"] = self.wet_surface_area
        hydrostatics["disp_volumes"] = self.volumes
        hydrostatics["disp_volume"] = self.volume
        hydrostatics["disp_mass"] = self.disp_mass(rho=rho)
        hydrostatics["center_of_buoyancy"] = self.center_of_buoyancy
        hydrostatics["waterplane_center"] = np.append(self.waterplane_center, 0.0)
        hydrostatics["waterplane_area"] = self.waterplane_area
        hydrostatics["transversal_metacentric_radius"] = self.transversal_metacentric_radius
        hydrostatics["longitudinal_metacentric_radius"] = self.longitudinal_metacentric_radius
        hydrostatics["transversal_metacentric_height"] = self.transversal_metacentric_height
        hydrostatics["longitudinal_metacentric_height"] = self.longitudinal_metacentric_height
        hydrostatics["hydrostatic_stiffness"] = self.compute_hydrostatic_stiffness(
            divergence=divergence, rho=rho, g=g)

        hydrostatics["length_overall"] = full_length
        hydrostatics["breadth_overall"] = full_breadth
        hydrostatics["depth"] = depth
        hydrostatics["draught"] = np.abs(coord_min[2])
        hydrostatics["length_at_waterline"] = wl_length
        hydrostatics["breadth_at_waterline"] = wl_breadth
        hydrostatics["length_overall_submerged"] = sub_length
        hydrostatics["breadth_overall_submerged"] = sub_breadth
        hydrostatics["inertia_matrix"] = self.compute_rigid_body_inertia(rho=rho)

        return hydrostatics


    ###################
    # Transformations #
    ###################

    def __add__(self, body_to_add: 'FloatingBody') -> 'FloatingBody':
        return self.join_bodies(body_to_add)

    def join_bodies(*bodies, name=None) -> 'FloatingBody':
        if name is None:
            name = "+".join(body.name for body in bodies)
        meshes = CollectionOfMeshes([body.mesh for body in bodies], name=f"{name}_mesh")
        dofs = FloatingBody.combine_dofs(bodies)

        if all(body.mass is not None for body in bodies):
            new_mass = sum(body.mass is not None for body in bodies)
        else:
            new_mass = None

        if (all(body.mass is not None for body in bodies)
                and all(body.center_of_mass is not None for body in bodies)):
            new_cog = sum(body.mass*body.center_of_mass is not None for body in bodies)/new_mass
        else:
            new_cog = None

        return FloatingBody(
            mesh=meshes, dofs=dofs, mass=new_mass, center_of_mass=new_cog, name=name
            )

    @staticmethod
    def combine_dofs(bodies) -> dict:
        """Combine the degrees of freedom of several bodies."""
        dofs = {}
        cum_nb_faces = accumulate(chain([0], (body.mesh.nb_faces for body in bodies)))
        total_nb_faces = sum(body.mesh.nb_faces for body in bodies)
        for body, nbf in zip(bodies, cum_nb_faces):
            # nbf is the cumulative number of faces of the previous subbodies,
            # that is the offset of the indices of the faces of the current body.
            for name, dof in body.dofs.items():
                new_dof = np.zeros((total_nb_faces, 3))
                new_dof[nbf:nbf+len(dof), :] = dof
                if '__' not in name:
                    new_dof_name = '__'.join([body.name, name])
                else:
                    # The body is probably a combination of bodies already.
                    # So for the associativity of the + operation,
                    # it is better to keep the same name.
                    new_dof_name = name
                dofs[new_dof_name] = new_dof
        return dofs

    def copy(self, name=None) -> 'FloatingBody':
        """Return a deep copy of the body.

        Parameters
        ----------
        name : str, optional
            a name for the new copy
        """
        new_body = copy.deepcopy(self)
        if name is None:
            new_body.name = f"copy_of_{self.name}"
            LOG.debug(f"Copy {self.name}.")
        else:
            new_body.name = name
            LOG.debug(f"Copy {self.name} under the name {name}.")
        return new_body

    def assemble_regular_array(self, distance, nb_bodies):
        """Create an regular array of identical bodies.

        Parameters
        ----------
        distance : float
            Center-to-center distance between objects in the array
        nb_bodies : couple of ints
            Number of objects in the x and y directions.

        Returns
        -------
        FloatingBody
        """
        bodies = (self.translated((i*distance, j*distance, 0), name=f"{i}_{j}") for j in range(nb_bodies[1]) for i in range(nb_bodies[0]))
        array = FloatingBody.join_bodies(*bodies)
        array.mesh = build_regular_array_of_meshes(self.mesh, distance, nb_bodies)
        array.name = f"array_of_{self.name}"
        return array

    def assemble_arbitrary_array(self, locations:np.ndarray):

        if not isinstance(locations, np.ndarray):
            raise TypeError('locations must be of type np.ndarray')
        assert locations.shape[1] == 2, 'locations must be of shape nx2, received {:}'.format(locations.shape)
        n = locations.shape[0]

        fb_list = []
        for idx, li in enumerate(locations):
            fb1 = self.copy()
            fb1.translate(np.append(li,0))
            fb1.name = 'arbitrary_array_body{:02d}'.format(idx)
            fb_list.append(fb1)

        arbitrary_array = fb_list[0].join_bodies(*fb_list[1:])

        return arbitrary_array

    def extract_faces(self, id_faces_to_extract, return_index=False):
        """Create a new FloatingBody by extracting some faces from the mesh.
        The dofs evolve accordingly.
        """
        if isinstance(self.mesh, CollectionOfMeshes):
            raise NotImplementedError  # TODO

        if return_index:
            new_mesh, id_v = Mesh.extract_faces(self.mesh, id_faces_to_extract, return_index)
        else:
            new_mesh = Mesh.extract_faces(self.mesh, id_faces_to_extract, return_index)
        new_body = FloatingBody(new_mesh)
        LOG.info(f"Extract floating body from {self.name}.")

        new_body.dofs = {}
        for name, dof in self.dofs.items():
            new_body.dofs[name] = dof[id_faces_to_extract, :]

        if return_index:
            return new_body, id_v
        else:
            return new_body

    def sliced_by_plane(self, plane):
        return FloatingBody(mesh=self.mesh.sliced_by_plane(plane), dofs=self.dofs, name=self.name)

    def minced(self, nb_slices=(8, 8, 4)):
        """Experimental method decomposing the mesh as a hierarchical structure.

        Parameters
        ----------
        nb_slices: Tuple[int, int, int]
            The number of slices in each of the x, y and z directions.
            Only powers of 2 are supported at the moment.

        Returns
        -------
        FloatingBody
        """
        minced_body = self.copy()

        # Extreme points of the mesh in each directions.
        x_min, x_max, y_min, y_max, z_min, z_max = self.mesh.axis_aligned_bbox
        sizes = [(x_min, x_max), (y_min, y_max), (z_min, z_max)]

        directions = [np.array(d) for d in [(1, 0, 0), (0, 1, 0), (0, 0, 1)]]

        def _slice_positions_at_depth(i):
            """Helper function.

            Returns a list of floats as follows:
            i=1 -> [1/2]
            i=2 -> [1/4, 3/4]
            i=3 -> [1/8, 3/8, 5/8, 7/8]
                   ...
            """
            denominator = 2**i
            return [numerator/denominator for numerator in range(1, denominator, 2)]

        # GENERATE ALL THE PLANES THAT WILL BE USED TO MINCE THE MESH
        planes = []
        for direction, nb_slices_in_dir, (min_coord, max_coord) in zip(directions, nb_slices, sizes):
            planes_in_dir = []

            depth_of_treelike_structure = int(np.log2(nb_slices_in_dir))
            for i_depth in range(1, depth_of_treelike_structure+1):
                planes_in_dir_at_depth = []
                for relative_position in _slice_positions_at_depth(i_depth):
                    slice_position = (min_coord + relative_position*(max_coord-min_coord))*direction
                    plane = Plane(normal=direction, point=slice_position)
                    planes_in_dir_at_depth.append(plane)
                planes_in_dir.append(planes_in_dir_at_depth)
            planes.append(planes_in_dir)

        # SLICE THE MESH
        intermingled_x_y_z = chain.from_iterable(zip_longest(*planes))
        for planes in intermingled_x_y_z:
            if planes is not None:
                for plane in planes:
                    minced_body = minced_body.sliced_by_plane(plane)
        return minced_body

    @inplace_transformation
    def mirror(self, plane):
        self.mesh.mirror(plane)
        for dof in self.dofs:
            self.dofs[dof] -= 2 * np.outer(np.dot(self.dofs[dof], plane.normal), plane.normal)
        for point_attr in ('geometric_center', 'rotation_center', 'center_of_mass'):
            if point_attr in self.__dict__ and self.__dict__[point_attr] is not None:
                self.__dict__[point_attr] -= 2 * (np.dot(self.__dict__[point_attr], plane.normal) - plane.c) * plane.normal
        return self

    @inplace_transformation
    def translate(self, *args):
        self.mesh.translate(*args)
        for point_attr in ('geometric_center', 'rotation_center', 'center_of_mass'):
            if point_attr in self.__dict__ and self.__dict__[point_attr] is not None:
                self.__dict__[point_attr] += args[0]
        return self

    @inplace_transformation
    def rotate(self, axis, angle):
        self.mesh.rotate(axis, angle)
        for point_attr in ('geometric_center', 'rotation_center', 'center_of_mass'):
            if point_attr in self.__dict__ and self.__dict__[point_attr] is not None:
                self.__dict__[point_attr] = axis.rotate_points([self.__dict__[point_attr]], angle)
        for dof in self.dofs:
            self.dofs[dof] = axis.rotate_vectors(self.dofs[dof], angle)
        return self

    @inplace_transformation
    def clip(self, plane):
        # Keep of copy of the full mesh
        if self.full_body is None:
            self.full_body = self.copy()

        # Clip mesh
        LOG.info(f"Clipping {self.name} with respect to {plane}")
        self.mesh.clip(plane)

        # Clip dofs
        ids = self.mesh._clipping_data['faces_ids']
        for dof in self.dofs:
            if len(ids) > 0:
                self.dofs[dof] = np.array(self.dofs[dof])[ids]
            else:
                self.dofs[dof] = np.empty((0, 3))
        return self

    def clipped(self, plane, **kwargs):
        # Same API as for the other transformations
        return self.clip(plane, inplace=False, **kwargs)

    @inplace_transformation
    def keep_immersed_part(self, free_surface=0.0, sea_bottom=-np.infty):
        """Remove the parts of the mesh above the sea bottom and below the free surface."""
        self.clip(Plane(normal=(0, 0, 1), point=(0, 0, free_surface)))
        if sea_bottom > -np.infty:
            self.clip(Plane(normal=(0, 0, -1), point=(0, 0, sea_bottom)))
        return self

    #############
    #  Display  #
    #############

    def __str__(self):
        return self.name

    def __repr__(self):
        return (f"{self.__class__.__name__}(mesh={self.mesh.name}, "
                f"dofs={{{', '.join(self.dofs.keys())}}}, name={self.name})")

    def show(self, **kwargs):
        from capytaine.ui.vtk.body_viewer import FloatingBodyViewer
        viewer = FloatingBodyViewer()
        viewer.add_body(self, **kwargs)
        viewer.show()
        viewer.finalize()

    def show_matplotlib(self, *args, **kwargs):
        return self.mesh.show_matplotlib(*args, **kwargs)

    def animate(self, motion, *args, **kwargs):
        """Display a motion as a 3D animation.

        Parameters
        ==========
        motion: dict or pd.Series or str
            A dict or series mapping the name of the dofs to its amplitude.
            If a single string is passed, it is assumed to be the name of a dof
            and this dof with a unit amplitude will be displayed.
        """
        from capytaine.ui.vtk.animation import Animation
        if isinstance(motion, str):
            motion = {motion: 1.0}
        elif isinstance(motion, xr.DataArray):
            motion = {k: motion.sel(radiating_dof=k).data for k in motion.coords["radiating_dof"].data}

        if any(dof not in self.dofs for dof in motion):
            missing_dofs = set(motion.keys()) - set(self.dofs.keys())
            raise ValueError(f"Trying to animate the body {self.name} using dof(s) {missing_dofs}, but no dof of this name is defined for {self.name}.")

        animation = Animation(*args, **kwargs)
        animation._add_actor(self.mesh.merged(), faces_motion=sum(motion[dof_name] * dof for dof_name, dof in self.dofs.items() if dof_name in motion))
        return animation




#!/usr/bin/env python
# coding: utf-8
"""Tool to import optional dependencies. Inspired by similar code in pandas."""

import importlib

def import_optional_dependency(module_name: str, package_name: str = None):
    try:
        module = importlib.import_module(module_name)
    except ImportError:
        if package_name is None:
            package_name = module_name

        message = (
            f"Missing optional dependency '{module_name}'. "
            f"Use pip or conda to install {package_name}."
        )
        raise ImportError(message) from None

    return module

def silently_import_optional_dependency(module_name: str):
    # Same as above, except it does not raise a exception when the module is not found.
    # Instead, simply returns None.
    try:
        module = importlib.import_module(module_name)
    except ImportError:
        module = None
    return module



#!/usr/bin/env python
# coding: utf-8

from collections import OrderedDict
from functools import wraps

def delete_first_lru_cache(maxsize=1):
    """Behaves like functools.lru_cache(), but the oldest data in the cache is
    deleted *before* computing a new one, in order to limit RAM usage when
    stored objects are big."""

    def decorator(f):
        cache = OrderedDict()

        @wraps(f)
        def decorated_f(*args, **kwargs):
            # /!\ cache only args

            if args in cache:
                # Get item in cache
                return cache[args]

            if len(cache) + 1 > maxsize:
                # Drop oldest item in cache.
                cache.popitem(last=False)

            # Compute and store
            result = f(*args, **kwargs)
            cache[args] = result

            return result

        return decorated_f

    return decorator


# if __name__ == "__main__":
#     import numpy as np
#     from functools import lru_cache

#     # @lru_cache(maxsize=1)
#     @delete_first_lru_cache(maxsize=1)
#     def f(i):
#         print(i)
#         np.random.seed(i)
#         a = np.random.random((20_000, 20_000))
#         print(a.nbytes / 2**30, "Go")
#         return a

#     print(f(1)[0, 0])
#     print(f(1)[0, 0])
#     print(f(2)[0, 0])
#     print(f(1)[0, 0])
#     print(f(3)[0, 0])



#!/usr/bin/env python
# coding: utf-8

from capytaine.ui.vtk.mesh_viewer import MeshViewer
from capytaine.ui.vtk.body_viewer import FloatingBodyViewer
from capytaine.ui.vtk.animation import Animation



#!/usr/bin/env python
# coding: utf-8
"""Computing the potential and velocity of Airy wave."""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import numpy as np


def airy_waves_potential(points, pb, convention="Nemoh"):
    """Compute the potential for Airy waves at a given point (or array of points).

    Parameters
    ----------
    points: array of shape (3) or (N x 3)
        coordinates of the points in which to evaluate the potential.
    pb: DiffractionProblem
        problem with the environmental conditions (g, rho, ...) of interest
    convention: str, optional
        convention for the incoming wave field. Accepted values: "Nemoh", "WAMIT".

    Returns
    -------
    array of shape (1) or (N x 1)
        The potential
    """
    assert convention.lower() in ["nemoh", "wamit"], \
        "Convention for wave field should be either Nemoh or WAMIT."

    x, y, z = points.T
    k = pb.wavenumber
    h = pb.depth
    wbar = x * np.cos(pb.wave_direction) + y * np.sin(pb.wave_direction)

    if 0 <= k*h < 20:
        cih = np.cosh(k*(z+h))/np.cosh(k*h)
        # sih = np.sinh(k*(z+h))/np.cosh(k*h)
    else:
        cih = np.exp(k*z)
        # sih = np.exp(k*z)

    if convention.lower() == "wamit":
        return  1j*pb.g/pb.omega * cih * np.exp(-1j * k * wbar)
    else:
        return -1j*pb.g/pb.omega * cih * np.exp(1j * k * wbar)


def airy_waves_velocity(points, pb, convention="Nemoh"):
    """Compute the fluid velocity for Airy waves at a given point (or array of points).

    Parameters
    ----------
    points: array of shape (3) or (N x 3)
        coordinates of the points in which to evaluate the potential.
    pb: DiffractionProblem
        problem with the environmental conditions (g, rho, ...) of interest
    convention: str, optional
        convention for the incoming wave field. Accepted values: "Nemoh", "WAMIT".

    Returns
    -------
    array of shape (3) or (N x 3)
        the velocity vectors
    """
    assert convention.lower() in ["nemoh", "wamit"], \
        "Convention for wave field should be either Nemoh or WAMIT."

    x, y, z = points.T
    k = pb.wavenumber
    h = pb.depth

    wbar = x * np.cos(pb.wave_direction) + y * np.sin(pb.wave_direction)

    if 0 <= k*h < 20:
        cih = np.cosh(k*(z+h))/np.cosh(k*h)
        sih = np.sinh(k*(z+h))/np.cosh(k*h)
    else:
        cih = np.exp(k*z)
        sih = np.exp(k*z)

    v = pb.g*k/pb.omega * \
        np.exp(1j * k * wbar) * \
        np.array([np.cos(pb.wave_direction) * cih, np.sin(pb.wave_direction) * cih, -1j * sih])

    if convention.lower() == "wamit":
        return np.conjugate(v.T)
    else:
        return v.T


def froude_krylov_force(pb, convention="Nemoh"):
    pressure = 1j * pb.omega * pb.rho * airy_waves_potential(pb.body.mesh.faces_centers, pb, convention=convention)
    forces = {}
    for dof in pb.influenced_dofs:
        # Scalar product on each face:
        normal_dof_amplitude_on_face = - np.sum(pb.body.dofs[dof] * pb.body.mesh.faces_normals, axis=1)
        # The minus sign in the above line is because we want the force of the fluid on the body and not the force of the body on the fluid.
        # Sum over all faces:
        forces[dof] = np.sum(pressure * normal_dof_amplitude_on_face * pb.body.mesh.faces_areas)
    return forces



#!/usr/bin/env python
# coding: utf-8
"""Prony decomposition: tool to approximate a function as a sum of exponentials.
Used in particular in the finite depth Green function.
"""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging

import numpy as np
from numpy.polynomial import polynomial
from scipy.optimize import curve_fit
from scipy.linalg import toeplitz

LOG = logging.getLogger(__name__)


def exponential_decomposition(X, F, m):
    """Use Prony's method to approximate the sampled real function F=f(X) as a sum of m
    exponential functions x → Σ a_i exp(lamda_i x).

    Parameters
    ----------
    X: 1D array
        sampling points.
    F: 1D array (same size as X)
        values of the function to approximate at the points of x.
    m: integer
        number of exponential functions

    Return
    ------
    a: 1D array (size m)
        coefficients of the exponentials
    lamda: 1D array (size m)
        growth rate of the exponentials
    """
    assert X.shape == F.shape

    # Compute the coefficients of the polynomials of Prony's method
    A = toeplitz(c=F[m-1:-1], r=F[:m][::-1])
    P, *_ = np.linalg.lstsq(A, F[m:], rcond=None)

    # Build and solve polynomial function
    coeffs = np.ones(m+1)
    # coeffs[:m] = -P[::-1]
    for i in range(m):
        coeffs[m-i-1] = -P[i]
    roots = polynomial.polyroots(coeffs)

    # Discard values where log is undefined
    roots = roots[np.logical_or(np.imag(roots) != 0.0, np.real(roots) >= 0.0)]

    # Deduce lamda and keep only interesting values
    lamda = np.real(np.log(roots)/(X[1] - X[0]))
    lamda = np.unique(lamda)
    lamda = lamda[np.logical_and(-20.0 < lamda, lamda < 0.0)]

    # Fit the values of 'a' on the curve
    def f(x, *ar):
        ar = np.asarray(ar)[:, np.newaxis]
        la = lamda[:, np.newaxis]
        return np.sum(ar * np.exp(la * x), axis=0)
    a, *_ = curve_fit(f, X, F, p0=np.zeros(lamda.shape))

    return a, lamda


def error_exponential_decomposition(X, F, a, lamda):
    """Compare exponential decomposition defined by the coefficients a and lamda to the reference
    values in F.

    Parameters
    ----------
    X: 1D array
        sampling points
    F: 1D array (same size as X)
        reference values
    a: 1D array
        coefficients of the exponentials
    lamda: 1D array (same size as a)
        growth rate of the exponentials

    Returns
    -------
    error: float
        mean square error of the decomposition
    """
    a = np.asarray(a)[:, np.newaxis]
    lamda = np.asarray(lamda)[:, np.newaxis]

    def f(x):
        return np.sum(a * np.exp(lamda*x), axis=0)

    return np.square(f(X) - F).mean()




#!/usr/bin/env python
#  -*- coding: utf-8 -*-
"""Tools to describe geometric objects in 3D.
Based on meshmagick <https://github.com/LHEEA/meshmagick> by François Rongère.
"""
# Copyright (C) 2017-2019 Matthieu Ancellin, based on the work of François Rongère
# See LICENSE file at <https://github.com/mancellin/capytaine>

from abc import ABC, abstractmethod

import numpy as np

e_x = np.array((1, 0, 0))
e_y = np.array((0, 1, 0))
e_z = np.array((0, 0, 1))


###########################################
#  DECORATOR FOR INPLACE TRANSFORMATIONS  #
###########################################

def inplace_transformation(inplace_function):
    """Decorator for methods transforming 3D objects:
    * Add the optional argument `inplace` to return a new object instead of doing the transformation in place.
    * If the object has properties cached in an "__internals__" dict, they are deleted.
    """
    def enhanced_inplace_function(self, *args, inplace=True, name=None, **kwargs):
        if not inplace:
            object3d = self.copy(name=name)
        else:
            object3d = self
        inplace_function(object3d, *args, **kwargs)
        if hasattr(object3d, '__internals__'):
            object3d.__internals__.clear()
        return object3d
    return enhanced_inplace_function


##############################
#  ABSTRACT 3D OBJECT CLASS  #
##############################

class Abstract3DObject(ABC):
    """Abstract class for 3d objects that can be transformed in 3d.
    The child classes have to define `mirror`, `rotate` and `translate`,
    then more routines such as `translate_x` and `translated` are automatically available."""

    @abstractmethod
    def translate(self, vector):
        pass

    @abstractmethod
    def rotate(self, axis, angle):
        pass

    @abstractmethod
    def mirror(self, plane):
        pass

    @inplace_transformation
    def translate_x(self, tx):
        return self.translate((tx, 0., 0.))

    @inplace_transformation
    def translate_y(self, ty):
        return self.translate((0., ty, 0.))

    @inplace_transformation
    def translate_z(self, tz):
        return self.translate((0., 0., tz))

    @inplace_transformation
    def translate_point_to_point(self, point_a, point_b):
        return self.translate(np.asarray(point_b) - np.asarray(point_a))

    @inplace_transformation
    def rotate_x(self, thetax):
        return self.rotate(Ox_axis, thetax)

    @inplace_transformation
    def rotate_y(self, thetay):
        return self.rotate(Oy_axis, thetay)

    @inplace_transformation
    def rotate_z(self, thetaz):
        return self.rotate(Oz_axis, thetaz)

    @inplace_transformation
    def rotate_around_center_to_align_vectors(self, center, vec1, vec2):
        """Rotate self such that if vec1 is in self, then it will point in the same direction as vec2."""
        vec1 = np.asarray(vec1)
        vec2 = np.asarray(vec2)
        if parallel_vectors_with_same_direction(vec1, vec2):
            return self
        else:
            if parallel_vectors(vec1, vec2):
                if parallel_vectors(vec1, e_x):
                    axis = Axis(vector=np.cross(vec1, e_y), point=center)
                else:
                    axis = Axis(vector=np.cross(vec1, e_x), point=center)
                return self.rotate(axis, np.pi)
            else:
                axis = Axis(vector=np.cross(vec1, vec2), point=center)
                return self.rotate(axis, np.arccos(np.dot(vec1, vec2)))

    def translated(self, *args, **kwargs):
        return self.translate(*args, inplace=False, **kwargs)

    def rotated(self, *args, **kwargs):
        return self.rotate(*args, inplace=False, **kwargs)

    def mirrored(self, *args, **kwargs):
        return self.mirror(*args, inplace=False, **kwargs)

    def translated_x(self, *args, **kwargs):
        return self.translate_x(*args, inplace=False, **kwargs)

    def translated_y(self, *args, **kwargs):
        return self.translate_y(*args, inplace=False, **kwargs)

    def translated_z(self, *args, **kwargs):
        return self.translate_z(*args, inplace=False, **kwargs)

    def translated_point_to_point(self, *args, **kwargs):
        return self.translate_point_to_point(*args, inplace=False, **kwargs)

    def rotated_x(self, *args, **kwargs):
        return self.rotate_x(*args, inplace=False, **kwargs)

    def rotated_y(self, *args, **kwargs):
        return self.rotate_y(*args, inplace=False, **kwargs)

    def rotated_z(self, *args, **kwargs):
        return self.rotate_z(*args, inplace=False, **kwargs)

    def rotated_around_center_to_align_vectors(self, *args, **kwargs):
        return self.rotate_around_center_to_align_vectors(*args, inplace=False, **kwargs)


######################
#  HELPER FUNCTIONS  #
######################

def orthogonal_vectors(vec1, vec2) -> bool:
    return np.linalg.norm(vec1 @ vec2) < 1e-6


def parallel_vectors(vec1, vec2) -> bool:
    return np.linalg.norm(np.cross(vec1, vec2)) < 1e-6


def parallel_vectors_with_same_direction(vec1, vec2) -> bool:
    return parallel_vectors(vec1, vec2) and np.dot(vec1, vec2) > 0


################
#  AXIS CLASS  #
################

class Axis(Abstract3DObject):
    def __init__(self, vector=(1, 0, 0), point=(0, 0, 0)):
        assert len(vector) == 3, "Vector of an axis should be given as a 3-ple of values."
        assert len(point) == 3, "Point of an axis should be given as a 3-ple of values."
        vector = np.array(vector, float)
        self.vector = vector / np.linalg.norm(vector)
        self.point = np.array(point, float)

    def __repr__(self):
        return f"Axis(vector={self.vector}, point={self.point})"

    def __contains__(self, other_point):
        if len(other_point) == 3:
            other_point = np.asarray(other_point, dtype=float)
            return parallel_vectors(other_point - self.point, self.vector)
        else:
            raise NotImplementedError

    def __eq__(self, other):
        if isinstance(self, Axis):
            return (self is other) or (self.point in other and parallel_vectors(self.vector, other.vector))
        else:
            return NotImplemented

    def is_orthogonal_to(self, other):
        if isinstance(other, Plane):
            return parallel_vectors(self.vector, other.normal)
        elif len(other) == 3:  # The other is supposed to be a vector given as a 3-ple
            return orthogonal_vectors(self.vector, other)
        else:
            raise NotImplementedError

    def is_parallel_to(self, other):
        if isinstance(other, Plane):
            return orthogonal_vectors(self.vector, other.normal)
        elif isinstance(other, Axis):
            return parallel_vectors(self.vector, other.vector)
        elif len(other) == 3:  # The other is supposed to be a vector given as a 3-ple
            return parallel_vectors(self.vector, other)
        else:
            raise NotImplementedError

    def angle_with_respect_to(self, other_axis: 'Axis') -> float:
        """Angle between two axes."""
        return np.arccos(np.dot(self.vector, other_axis.vector))

    ################################
    #  Transformation of the axis  #
    ################################

    def copy(self, name=None):
        return Axis(vector=self.vector.copy(), point=self.point.copy())

    @inplace_transformation
    def translate(self, vector):
        self.point += vector
        return self

    @inplace_transformation
    def rotate(self, axis, angle):
        rot_matrix = axis.rotation_matrix(angle)
        self.point = rot_matrix @ self.point
        self.vector = rot_matrix @ self.vector
        return self

    @inplace_transformation
    def mirror(self, plane):
        self.point -= 2 * (self.point @ plane.normal - plane.c) * plane.normal
        self.vector -= 2 * (self.vector @ plane.normal) * plane.normal
        return self

    ###########
    #  Other  #
    ###########

    def rotation_matrix(self, theta):
        """Rotation matrix around the vector according to Rodrigues' formula."""
        ux, uy, uz = self.vector
        W = np.array([[0, -uz, uy],
                      [uz, 0, -ux],
                      [-uy, ux, 0]])
        return np.identity(3) + np.sin(theta)*W + 2*np.sin(theta/2)**2 * (W @ W)

    def rotate_vectors(self, vectors, angle):
        vectors = np.asarray(vectors)
        return (self.rotation_matrix(angle) @ vectors.T).T

    def rotate_points(self, points, angle):
        points = np.asarray(points)
        return self.rotate_vectors(points - self.point, angle) + self.point


Ox_axis = Axis(vector=e_x, point=(0, 0, 0))
Oy_axis = Axis(vector=e_y, point=(0, 0, 0))
Oz_axis = Axis(vector=e_z, point=(0, 0, 0))


#################
#  PLANE CLASS  #
#################

class Plane(Abstract3DObject):
    """3D plane, oriented by the direction of their normal."""
    def __init__(self, normal=(0.0, 0.0, 1.0), point=(0.0, 0.0, 0.0)):
        normal = np.asarray(normal, dtype=float)
        self.normal = normal / np.linalg.norm(normal)
        self.point = np.asarray(point, dtype=float)

    def __repr__(self):
        return f"Plane(normal={self.normal}, point={self.point})"

    def __contains__(self, other):
        if isinstance(other, Axis):
            return other.point in self and orthogonal_vectors(self.normal, other.vector)
        elif len(other) == 3:
            return orthogonal_vectors(other - self.point, self.normal)
        else:
            raise NotImplementedError

    def __eq__(self, other):
        """Plane are considered equal only when their normal are pointing in the same direction."""
        if isinstance(other, Plane):
            return ((self is other) or
                    (other.point in self and parallel_vectors_with_same_direction(self.normal, other.normal)))
        else:
            return NotImplemented

    def is_orthogonal_to(self, other):
        if isinstance(other, Axis):
            return parallel_vectors(self.normal, other.vector)
        elif isinstance(other, Plane):
            return orthogonal_vectors(self.normal, other.normal)
        elif len(other) == 3:  # The other is supposed to be a vector given as a 3-ple
            return parallel_vectors(self.normal, other)
        else:
            raise NotImplementedError

    @property
    def c(self):
        """Distance from plane to origin."""
        return np.linalg.norm(self.normal @ self.point)

    #################################
    #  Transformation of the plane  #
    #################################

    def copy(self, name=None):
        return Plane(normal=self.normal.copy(), point=self.point.copy())

    @inplace_transformation
    def translate(self, vector):
        self.point = self.point + np.asarray(vector)
        return self

    @inplace_transformation
    def rotate(self, axis, angle):
        rot_matrix = axis.rotation_matrix(angle)
        self.point = rot_matrix @ self.point
        self.normal = rot_matrix @ self.normal
        return self

    @inplace_transformation
    def mirror(self, plane):
        self.point -= 2 * (self.point @ plane.normal - plane.c) * plane.normal
        self.normal -= 2 * (self.normal @ plane.normal) * plane.normal
        return self

    ###########
    #  Other  #
    ###########

    def distance_to_point(self, points):
        """
        Return the orthogonal distance of points with respect to the plane.
        The distance is counted positively on one side of the plane and negatively on the other.

        Parameters
        ----------
        points : ndarray
            Array of points coordinates

        Returns
        -------
        dist : ndarray
            Array of distances of points with respect to the plane
        """
        return np.dot(points, self.normal) - np.dot(self.point, self.normal)

    def get_edge_intersection(self, p0, p1):
        """
        Returns the coordinates of the intersection point between the plane and the edge P0P1.

        Parameters
        ----------
        p0 : ndarray
            Coordinates of point p0
        p1 : ndarray
            Coordinates of point P1

        Returns
        -------
        I : ndarray
            Coordinates of intersection point
        """
        assert len(p0) == 3 and len(p1) == 3

        p0n = np.dot(p0, self.normal)
        p1n = np.dot(p1, self.normal)
        t = (p0n - self.c) / (p0n - p1n)
        if t < 0. or t > 1.:
            raise RuntimeError('Intersection is outside the edge')
        return (1-t) * p0 + t * p1


yOz_Plane = Plane(normal=e_x, point=(0, 0, 0))
xOz_Plane = Plane(normal=e_y, point=(0, 0, 0))
xOy_Plane = Plane(normal=e_z, point=(0, 0, 0))




#!/usr/bin/env python
# coding: utf-8
"""Generate meshes of cylinders and disks"""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging
from itertools import product

import numpy as np

from capytaine.meshes.geometry import xOy_Plane, xOz_Plane, yOz_Plane, e_x, e_z, Oz_axis
from capytaine.meshes.meshes import Mesh
from capytaine.meshes.collections import CollectionOfMeshes
from capytaine.meshes.symmetric import TranslationalSymmetricMesh, AxialSymmetricMesh, ReflectionSymmetricMesh
from capytaine.bodies.bodies import FloatingBody

LOG = logging.getLogger(__name__)


##########
#  Disk  #
##########

class Disk(FloatingBody):
    """(One-sided) disk.

    Parameters
    ----------
    radius : float, optional
        radius of the disk
    resolution : 2-ple of int, optional
        number of panels along a radius and around the disk
    center : 3-ple or array of shape (3,), optional
        position of the geometric center of the disk
    normal: 3-ple of floats, optional
        normal vector, default: along x axis
    axial_symmetry : bool, optional
        if True, use the axial symmetry to speed up the computations
    reflection_symmetry : bool, optional
        if True, use the reflection symmetry to speed up the computations
    name : str, optional
        a string naming the floating body
    """

    def __init__(self, radius=1.0, resolution=(3, 5),
                 center=(0, 0, 0), normal=(1, 0, 0),
                 reflection_symmetry=False, axial_symmetry=False,
                 name=None):
        assert radius > 0, "Radius of the disk mesh should be given as a positive value."

        assert len(resolution) == 2, "Resolution of a disk should be given as a couple a values."
        assert all([h > 0 for h in resolution]), "Resolution of the disk mesh should be given as positive values."
        assert all([i == int(i) for i in resolution]), "Resolution of a disk should be given as integer values."

        assert len(center) == 3, "Position of the center of a disk should be given a 3-ple of values."

        self.radius = float(radius)
        self.geometric_center = np.asarray(center, dtype=float)
        nr, ntheta = resolution

        if reflection_symmetry and ntheta % 2 == 1:
            raise ValueError("To use the reflection symmetry of the mesh, "
                             "it should have an even number of panels in this direction.")

        if reflection_symmetry and axial_symmetry:
            raise NotImplementedError("Disk generators with both symmetries have not been implemented.")

        if name is None:
            name = f"disk_{next(Mesh._ids)}"

        LOG.debug(f"New disk body of radius {radius} and resolution ({nr}, {ntheta}), named {name}.")

        if reflection_symmetry:
            half_mesh = Disk.generate_disk_mesh(radius=self.radius, theta_max=np.pi/2,
                                                nr=nr, ntheta=ntheta//2,
                                                name=f"half_of_{name}_mesh")
            mesh = ReflectionSymmetricMesh(half_mesh, plane=xOz_Plane, name=f"{name}_mesh")

        elif axial_symmetry:
            mesh_slice = Disk.generate_disk_mesh(radius=self.radius, theta_max=np.pi/ntheta,
                                                 nr=nr, ntheta=1,
                                                 name=f"slice_of_{name}_mesh")
            mesh_slice.rotate_around_center_to_align_vectors((0, 0, 0), e_x, e_z)  # Convoluted way to avoid a warning message in AxialSymmetry...
            mesh = AxialSymmetricMesh(mesh_slice, axis=Oz_axis, nb_repetitions=ntheta - 1, name=f"{name}_mesh")
            mesh.rotate_around_center_to_align_vectors((0, 0, 0), e_z, e_x)

        else:
            mesh = Disk.generate_disk_mesh(radius=self.radius, nr=nr, ntheta=ntheta, name=f"{name}_mesh")

        mesh.rotate_around_center_to_align_vectors((0, 0, 0), mesh.faces_normals[0], normal)
        mesh.translate(center)
        FloatingBody.__init__(self, mesh=mesh, name=name)

    @staticmethod
    def generate_disk_mesh(radius=1.0, theta_max=np.pi,
                           nr=2, ntheta=4,
                           center=(0, 0, 0), normal=(1, 0, 0),
                           name=None) -> Mesh:
        theta_range = np.linspace(0, 2*theta_max, ntheta+1)
        r_range = np.linspace(0.0, radius, nr+1)

        nodes = np.zeros(((ntheta+1)*(nr+1), 3), dtype=float)
        for i, (r, t) in enumerate(product(r_range, theta_range)):
            y = +r * np.sin(t)
            z = -r * np.cos(t)
            nodes[i, :] = (0, y, z)

        panels = np.zeros((ntheta*nr, 4), dtype=int)

        for k, (i, j) in enumerate(product(range(0, nr), range(0, ntheta))):
            panels[k, :] = (
                j+i*(ntheta+1),
                j+1+i*(ntheta+1),
                j+1+(i+1)*(ntheta+1),
                j+(i+1)*(ntheta+1)
            )

        mesh = Mesh(nodes, panels, name=name)
        mesh.merge_duplicates()
        mesh.heal_triangles()
        mesh.rotate_around_center_to_align_vectors((0, 0, 0), mesh.faces_normals[0], normal)
        mesh.translate(center)
        return mesh


##############
#  Cylinder  #
##############

class HorizontalCylinder(FloatingBody):
    """Horizontal cylinder

    Parameters
    ----------
    length : float, optional
        length of the cylinder
    radius : float, optional
        radius of the cylinder
    center : 3-ple or array of shape (3,), optional
        position of the geometric center of the cylinder
    nx : int, optional
        number of circular slices
    ntheta : int, optional
        number of panels along a circular slice of the cylinder
    nr : int, optional
        number of panels along a radius on the extremities of the cylinder
    reflection_symmetry : bool, optional
        if True, returns a ReflectionSymmetricMesh
    translation_symmetry : bool, optional
        if True, uses a TranslationalSymmetricMesh internally for the main part of the cylinder
    name : str, optional
        a string naming the floating body
    """

    def __init__(self, length=10.0, radius=1.0, center=(0, 0, 0),
                 nx=10, ntheta=10, nr=2,
                 reflection_symmetry=True, translation_symmetry=False,
                 clever=None,
                 name=None):
        self.length = length
        self.radius = radius
        self.geometric_center = np.asarray(center, dtype=float)

        if name is None:
            name = f"cylinder_{next(Mesh._ids)}"

        ntheta = 2*(ntheta//2)  # Temporary fix to avoid mismatch in mesh
        # When symmetries are used, one needs an even number of panels.
        # TODO: When symmetries are not used, implement the odd case.

        if clever is not None:
            LOG.warning("Deprecation warning: `clever` argument for HorizontalCylinder is deprecated."
                        "Use `reflection_symmetry` and/or `translation_symmetry` instead.")

        open_cylinder = self._generate_open_cylinder_mesh(nx, ntheta,
                                                          reflection_symmetry=reflection_symmetry,
                                                          translation_symmetry=translation_symmetry,
                                                          name=f"body_of_{name}")

        if nr == 0:  # No sides
            mesh = open_cylinder

        else:  # Sides
            side = Disk(radius=radius, center=(-np.array([length/2, 0, 0])), normal=(-1, 0, 0),
                        reflection_symmetry=reflection_symmetry,
                        resolution=(nr, ntheta), name=f"side_of_{name}").mesh

            other_side = side.copy(name=f"other_side_of_{name}_mesh")
            other_side.mirror(yOz_Plane)

            if reflection_symmetry:  # Knit the sides into the symmetric representation of the open cylinder
                half_sides = CollectionOfMeshes((side.half, other_side.half), name=f"half_sides_of_{name}_mesh")
                half_mesh = CollectionOfMeshes((open_cylinder.half, half_sides), name=f"half_{name}_mesh")
                mesh = ReflectionSymmetricMesh(half_mesh, plane=xOz_Plane, name=f"{name}_mesh")
            else:
                sides = CollectionOfMeshes((side, other_side), name=f"sides_of_cylinder_{name}_mesh")
                mesh = CollectionOfMeshes((open_cylinder, sides), name=f"{name}_mesh")

        if not reflection_symmetry and not translation_symmetry:
            mesh = mesh.merged()

        mesh.heal_mesh()

        mesh.translate(self.geometric_center)
        mesh.name = f"{name}_mesh"

        FloatingBody.__init__(self, mesh=mesh, name=name)

    def _generate_open_cylinder_mesh(self, nx, ntheta, reflection_symmetry, translation_symmetry, name=None):
        """Open horizontal cylinder using the symmetries (translation and reflection) to speed up the computations"""
        theta_max = np.pi
        theta = np.linspace(0, theta_max, ntheta//2+1)
        X = np.array([0, self.length/nx])

        # Nodes
        nodes = np.zeros(((ntheta//2+1)*2, 3), dtype=float)

        for i, (t, x) in enumerate(product(theta, X)):
            y = + self.radius * np.sin(t)
            z = - self.radius * np.cos(t)
            nodes[i, :] = (x, y, z)
        nodes += -np.array([self.length/2, 0, 0])

        # Connectivities
        panels = np.zeros((ntheta//2, 4), dtype=int)

        for k, i in enumerate(range(0, ntheta//2)):
            panels[k, :] = (2*i, 2*i+2, 2*i+3, 2*i+1)
        half_ring = Mesh(nodes, panels, name=f"half_ring_of_{name}_mesh")

        if reflection_symmetry:
            if nx == 1:
                half_cylinder = half_ring
            else:
                half_cylinder = TranslationalSymmetricMesh(half_ring, translation=np.asarray([self.length / nx, 0.0, 0.0]),
                                                           nb_repetitions=nx-1, name=f"half_{name}_mesh")
                if not translation_symmetry:
                    half_cylinder = half_cylinder.merged()

            return ReflectionSymmetricMesh(half_cylinder, plane=xOz_Plane, name=f"{name}_mesh")

        else:
            strip = half_ring + half_ring.mirrored(plane=xOz_Plane)
            if nx == 1:
                return strip

            else:
                cylinder = TranslationalSymmetricMesh(strip, translation=np.asarray([self.length / nx, 0.0, 0.0]),
                                                      nb_repetitions=nx-1, name=f"half_{name}_mesh")
                if not translation_symmetry:
                    cylinder = cylinder.merged()

            return cylinder


class VerticalCylinder(FloatingBody):
    """Vertical cylinder.

    Parameters
    ----------
    length : float, optional
        length of the cylinder
    radius : float, optional
        radius of the cylinder
    center : 3-ple or array of shape (3,), optional
        position of the geometric center of the cylinder
    nx : int, optional
        number of circular slices
    ntheta : int, optional
        number of panels along a circular slice of the cylinder
    nr : int, optional
        number of panels along a radius on the extremities of the cylinder
    clever : bool, optional
        if True, uses the mesh symmetries
    name : str, optional
        a string naming the floating body
    """

    def __init__(self, length=10.0, radius=1.0, center=(0, 0, 0),
                 nx=10, ntheta=10, nr=2,
                 clever=True, name=None):
        self.length = length
        self.radius = radius
        self.geometric_center = np.asarray(center, dtype=float)

        if name is None:
            name = f"cylinder_{next(Mesh._ids)}"

        open_cylinder = AxialSymmetricMesh.from_profile(
            lambda z: radius,
            z_range=np.linspace(-length/2, length/2, nx+1),
            nphi=ntheta)

        if nr > 0:
            top_side = Disk(radius=radius, center=(0, 0, length/2),
                            axial_symmetry=True, normal=(0, 0, 1),
                            resolution=(nr, ntheta), name=f"top_side_of_{name}").mesh

            bottom_side = top_side.copy(name=f"bottom_side_of_{name}_mesh")
            bottom_side.mirror(xOy_Plane)

            mesh = AxialSymmetricMesh.join_meshes(open_cylinder, top_side, bottom_side)
        else:
            mesh = open_cylinder

        if not clever:
            mesh = mesh.merged()
            mesh.merge_duplicates()
            mesh.heal_triangles()

        mesh.translate(center)
        mesh.name = f"{name}_mesh"

        FloatingBody.__init__(self, mesh=mesh, name=name)



#!/usr/bin/env python
# coding: utf-8
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

from .__about__ import (
    __title__, __description__, __version__, __author__, __uri__, __license__
)

from capytaine.meshes.geometry import Axis, Plane, xOz_Plane, yOz_Plane, xOy_Plane
from capytaine.meshes.meshes import Mesh
from capytaine.meshes.collections import CollectionOfMeshes
from capytaine.meshes.symmetric import ReflectionSymmetricMesh, TranslationalSymmetricMesh, AxialSymmetricMesh

from capytaine.bodies.bodies import FloatingBody
from capytaine.bodies.predefined.spheres import Sphere
from capytaine.bodies.predefined.cylinders import VerticalCylinder, HorizontalCylinder, Disk
from capytaine.bodies.predefined.rectangles import Rectangle, RectangularParallelepiped, OpenRectangularParallelepiped

from capytaine.bem.problems_and_results import RadiationProblem, DiffractionProblem
from capytaine.bem.solver import Nemoh, BEMSolver
from capytaine.bem.engines import BasicMatrixEngine, HierarchicalToeplitzMatrixEngine
from capytaine.green_functions.delhommeau import Delhommeau, XieDelhommeau

from capytaine.post_pro.free_surfaces import FreeSurface

from capytaine.io.xarray import assemble_dataset



#!/usr/bin/env python
# coding: utf-8
"""Generate spherical bodies."""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging

import numpy as np

from capytaine.meshes.geometry import Axis
from capytaine.meshes.meshes import Mesh
from capytaine.meshes.symmetric import AxialSymmetricMesh
from capytaine.bodies.bodies import FloatingBody

LOG = logging.getLogger(__name__)


class Sphere(FloatingBody):
    """Sphere

    Parameters
    ----------
    radius : float
        radius of the sphere
    center : 3-ple or array of shape (3,)
        position of the geometric center of the sphere
    ntheta : int
        number of panels along a meridian (or number of parallels-1)
    nphi : int
        number of panels along a parallel (or number of meridians-1)
    axial_symmetry : bool
        if True, use the axial symmetry to build the mesh (default: True)
    clip_free_surface : bool
        if True, only mesh the part of the sphere where z < 0 (default: False),
        can be used with center to obtain any clipped sphere,
        if True, then ntheta is the number of parallel below the free surface.
    name : string
        a name identifying the sphere (default: "sphere_id" where id is an unique integer).
    """

    def __init__(self, *, radius=1.0, center=(0, 0, 0),
                 ntheta=10, nphi=10, clip_free_surface=False,
                 axial_symmetry=True, clever=None,
                 name=None):
        self.radius = radius
        self.geometric_center = np.array(center, dtype=float)

        if clever is not None:
            LOG.warning("Deprecation warning: `clever` argument for Sphere is deprecated. "
                        "Use `axial_symmetry` instead.")

        if name is None:
            name = f"sphere_{next(Mesh._ids)}"

        mesh = self._generate_mesh_using_symmetry(ntheta, nphi, clip_free_surface, f"{name}_mesh")

        if not axial_symmetry:
            mesh = mesh.merged()
            mesh.merge_duplicates()
            mesh.heal_triangles()

        FloatingBody.__init__(self, mesh=mesh, name=name)


    def _generate_mesh_using_symmetry(self, ntheta, nphi, clip_free_surface, mesh_name):
        if clip_free_surface:
            if self.geometric_center[2] < -self.radius:  # fully immersed
                theta_max = np.pi
            elif self.geometric_center[2] < self.radius:
                theta_max = np.arccos(self.geometric_center[2]/self.radius)
            else:
                raise ValueError("Impossible to mesh the immersed hull of a sphere completely out of the water")
        else:
            theta_max = np.pi

        theta = np.linspace(0.0, theta_max, ntheta+1)
        points_on_a_meridian = (
                self.radius * np.stack([np.sin(theta), np.zeros_like(theta), -np.cos(theta)], axis=1)
                + self.geometric_center
                )

        symmetry_axis = Axis(vector=[0, 0, 1], point=self.geometric_center)
        return AxialSymmetricMesh.from_profile(points_on_a_meridian, axis=symmetry_axis, nphi=nphi, name=mesh_name)




#!/usr/bin/env python
# coding: utf-8
"""Computation of the Kochin function."""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import numpy as np


def compute_kochin(result, theta, ref_point=(0.0, 0.0)):
    """Compute the far field coefficient

    Parameters
    ----------
    result: LinearPotentialFlowResult
        solved potential flow problem
    theta: float or 1-dim array of floats
        angles at which the coefficient is computed
    ref_point: couple of float, optional
        point of reference around which the far field coefficient is computed

    Returns
    -------
    H: same type as theta
        values of the Kochin function
    """

    if result.sources is None:
        raise Exception(f"""The values of the sources of {result} cannot been found.
        They probably have not been stored by the solver because the option keep_details=True have not been set.
        Please re-run the resolution with this option.""")

    k = result.wavenumber
    h = result.depth

    # omega_bar.shape = (nb_faces, 2) @ (2, nb_theta)
    omega_bar = (result.body.mesh.faces_centers[:, 0:2] - ref_point) @ (np.cos(theta), np.sin(theta))

    if 0 <= k*h < 20:
        cih = np.cosh(k*(result.body.mesh.faces_centers[:, 2]+h))/np.cosh(k*h)
    else:
        cih = np.exp(k*result.body.mesh.faces_centers[:, 2])

    # cih.shape = (nb_faces,)
    # omega_bar.T.shape = (nb_theta, nb_faces)
    # result.body.mesh.faces_areas.shape = (nb_faces,)
    zs = cih * np.exp(-1j * k * omega_bar.T) * result.body.mesh.faces_areas

    # zs.shape = (nb_theta, nb_faces)
    # result.sources.shape = (nb_faces,)
    return zs @ result.sources/(4*np.pi)




#!/usr/bin/env python
# coding: utf-8
"""Special meshes with symmetries, useful to speed up the computations."""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging
import reprlib
from typing import Union, Callable, Iterable

import numpy as np

from capytaine.meshes.meshes import Mesh
from capytaine.meshes.collections import CollectionOfMeshes
from capytaine.meshes.geometry import Axis, Plane, Oz_axis, inplace_transformation

LOG = logging.getLogger(__name__)


class SymmetricMesh(CollectionOfMeshes):
    def __repr__(self):
        reprer = reprlib.Repr()
        reprer.maxstring = 90
        reprer.maxother = 90
        slice_name = reprer.repr(self._meshes[0])
        if self.name is not None:
            return f"{self.__class__.__name__}({slice_name}, name={self.name})"
        else:
            return f"{self.__class__.__name__}({slice_name})"


class ReflectionSymmetricMesh(SymmetricMesh):
    """A mesh with one vertical symmetry plane.

    Parameters
    ----------
    half : Mesh or CollectionOfMeshes
        a mesh describing half of the body
    plane : Plane
        the symmetry plane across which the half body is mirrored
    name :str, optional
        a name for the mesh
    """

    def __init__(self, half: Union[Mesh, CollectionOfMeshes], plane: Plane, name=None):
        assert isinstance(half, Mesh) or isinstance(half, CollectionOfMeshes)
        assert isinstance(plane, Plane)
        assert plane.normal[2] == 0, "Only vertical reflection planes are supported in ReflectionSymmetry classes."

        other_half = half.mirrored(plane, name=f"mirrored_of_{str(half)}")

        super().__init__((half, other_half), name=name)

        self.plane = plane.copy()

        if self.name is not None:
            LOG.debug(f"New mirror symmetric mesh: {self.name}.")
        else:
            LOG.debug(f"New mirror symmetric mesh.")

    @property
    def half(self):
        return self[0]

    def tree_view(self, fold_symmetry=True, **kwargs):
        if fold_symmetry:
            return (str(self) + '\n' + ' ├─' + self.half.tree_view().replace('\n', '\n │ ') + '\n'
                    + f" └─mirrored copy of the above {str(self.half)}")
        else:
            return CollectionOfMeshes.tree_view(self, **kwargs)

    def __deepcopy__(self, *args):
        return ReflectionSymmetricMesh(self.half.copy(), self.plane, name=self.name)

    def join_meshes(*meshes, name=None):
        assert all(isinstance(mesh, ReflectionSymmetricMesh) for mesh in meshes), \
            "Only meshes with the same symmetry can be joined together."
        assert all(meshes[0].plane == mesh.plane for mesh in meshes), \
            "Only reflection symmetric meshes with the same reflection plane can be joined together."
        half_mesh = CollectionOfMeshes([mesh.half for mesh in meshes], name=f"half_of_{name}" if name is not None else None)
        return ReflectionSymmetricMesh(half_mesh, plane=meshes[0].plane, name=name)

    @inplace_transformation
    def translate(self, vector):
        self.plane.translate(vector)
        CollectionOfMeshes.translate(self, vector)
        return self

    @inplace_transformation
    def rotate(self, axis: Axis, angle: float):
        self.plane.rotate(axis, angle)
        CollectionOfMeshes.rotate(self, axis, angle)
        return self

    @inplace_transformation
    def mirror(self, plane: Plane):
        self.plane.mirror(plane)
        CollectionOfMeshes.mirror(self, plane)
        return self


class TranslationalSymmetricMesh(SymmetricMesh):
    """A mesh with a repeating pattern by translation.

    Parameters
    ----------
    mesh_slice : Mesh or CollectionOfMeshes
        the pattern that will be repeated to form the whole body
    translation : array(3)
        the vector of the translation
    nb_repetitions : int, optional
        the number of repetitions of the pattern (excluding the original one, default: 1)
    name : str, optional
        a name for the mesh
    """

    def __init__(self, mesh_slice: Union[Mesh, CollectionOfMeshes], translation, nb_repetitions=1, name=None):
        assert isinstance(mesh_slice, Mesh) or isinstance(mesh_slice, CollectionOfMeshes)
        assert isinstance(nb_repetitions, int)
        assert nb_repetitions >= 1

        translation = np.asarray(translation).copy()
        assert translation.shape == (3,)
        assert translation[2] == 0  # Only horizontal translation are supported.

        slices = [mesh_slice]
        for i in range(1, nb_repetitions+1):
            slices.append(mesh_slice.translated(vector=i*translation, name=f"repetition_{i}_of_{mesh_slice.name}"))

        super().__init__(slices, name=name)

        self.translation = translation

        if self.name is not None:
            LOG.debug(f"New translation symmetric mesh: {self.name}.")
        else:
            LOG.debug(f"New translation symmetric mesh.")

    @property
    def first_slice(self):
        return self[0]

    def tree_view(self, fold_symmetry=True, **kwargs):
        if fold_symmetry:
            return (str(self) + '\n' + ' ├─' + self.first_slice.tree_view().replace('\n', '\n │ ') + '\n'
                    + f" └─{len(self)-1} translated copies of the above {str(self.first_slice)}")
        else:
            return CollectionOfMeshes.tree_view(self, **kwargs)

    def __deepcopy__(self, *args):
        return TranslationalSymmetricMesh(self.first_slice.copy(), self.translation, nb_repetitions=len(self) - 1, name=self.name)

    @inplace_transformation
    def translate(self, vector):
        CollectionOfMeshes.translate(self, vector)
        return self

    @inplace_transformation
    def rotate(self, axis: Axis, angle: float):
        self.translation = axis.rotate_vector([self.translation], angle)[0, :]
        CollectionOfMeshes.rotate(self, axis, angle)
        return self

    @inplace_transformation
    def mirror(self, plane: Plane):
        self.translation -= 2 * (self.translation @ plane.normal) * plane.normal
        CollectionOfMeshes.mirror(self, plane)
        return self

    def join_meshes(*meshes, name=None):
        assert all(isinstance(mesh, TranslationalSymmetricMesh) for mesh in meshes), \
            "Only meshes with the same symmetry can be joined together."
        assert all(np.allclose(meshes[0].translation, mesh.translation) for mesh in meshes), \
            "Only translation symmetric meshes with the same translation vector can be joined together."
        assert all(len(meshes[0]) == len(mesh) for mesh in meshes), \
            "Only symmetric meshes with the same number of elements can be joined together."
        mesh_strip = CollectionOfMeshes([mesh.first_slice for mesh in meshes], name=f"strip_of_{name}" if name is not None else None)
        return TranslationalSymmetricMesh(mesh_strip, translation=meshes[0].translation, nb_repetitions=len(meshes[0]) - 1, name=name)


def build_regular_array_of_meshes(base_mesh, distance, nb_bodies):
    """Create an array of objects using TranslationalSymmetries.

    Parameters
    ----------
    base_mesh : Mesh or CollectionOfMeshes or SymmetricMesh
        The mesh to duplicate to create the array
    distance : float
        Center-to-center distance between objects in the array
    nb_bodies : couple of ints
        Number of objects in the x and y directions.

    Returns
    -------
    TranslationalSymmetricMesh
    """
    if nb_bodies[0] == 1:
        line = base_mesh
    else:
        line = TranslationalSymmetricMesh(base_mesh, translation=(distance, 0.0, 0.0), nb_repetitions=nb_bodies[0] - 1,
                                          name=f'line_of_{base_mesh.name}')
    if nb_bodies[1] == 1:
        array = line
    else:
        array = TranslationalSymmetricMesh(line, translation=(0.0, distance, 0.0), nb_repetitions=nb_bodies[1] - 1,
                                           name=f'array_of_{base_mesh.name}')
    return array


class AxialSymmetricMesh(SymmetricMesh):
    """A mesh with a repeating pattern by rotation.

    Parameters
    ----------
    mesh_slice : Mesh or CollectionOfMeshes
        the pattern that will be repeated to form the whole body
    axis : Axis, optional
        symmetry axis
    nb_repetitions : int, optional
        the number of repetitions of the pattern (excluding the original one, default: 1)
    name : str, optional
        a name for the mesh
    """
    def __init__(self, mesh_slice: Union[Mesh, CollectionOfMeshes], axis: Axis=Oz_axis, nb_repetitions: int=1, name=None):
        assert isinstance(mesh_slice, Mesh) or isinstance(mesh_slice, CollectionOfMeshes)
        assert isinstance(nb_repetitions, int)
        assert nb_repetitions >= 1
        assert isinstance(axis, Axis)

        slices = [mesh_slice]
        for i in range(1, nb_repetitions+1):
            slices.append(mesh_slice.rotated(axis, angle=2*i*np.pi/(nb_repetitions+1),
                                             name=f"rotation_{i}_of_{mesh_slice.name}"))

        super().__init__(slices, name=name)

        if not axis.is_parallel_to(Oz_axis):
            LOG.warning(f"{self.name} is an axi-symmetric mesh along a non vertical axis.")

        self.axis = axis.copy()

        if self.name is not None:
            LOG.debug(f"New rotation symmetric mesh: {self.name}.")
        else:
            LOG.debug(f"New rotation symmetric mesh.")

    @staticmethod
    def from_profile(profile: Union[Callable, Iterable[float]],
                     z_range: Iterable[float]=np.linspace(-5, 0, 20),
                     axis: Axis=Oz_axis,
                     nphi: int=20,
                     name=None):
        """Return a floating body using the axial symmetry.
        The shape of the body can be defined either with a function defining the profile as [f(z), 0, z] for z in z_range.
        Alternatively, the profile can be defined as a list of points.
        The number of vertices along the vertical direction is len(z_range) in the first case and profile.shape[0] in the second case.

        Parameters
        ----------
        profile : function(float → float)  or  array(N, 3)
            define the shape of the body either as a function or a list of points.
        z_range: array(N), optional
            used only if the profile is defined as a function.
        axis : Axis
            symmetry axis
        nphi : int, optional
            number of vertical slices forming the body
        name : str, optional
            name of the generated body (optional)

        Returns
        -------
        AxialSymmetricMesh
            the generated mesh
        """

        if name is None:
            name = "axisymmetric_mesh"

        if callable(profile):
            z_range = np.asarray(z_range)
            x_values = [profile(z) for z in z_range]
            profile_array = np.stack([x_values, np.zeros(len(z_range)), z_range]).T
        else:
            profile_array = np.asarray(profile)

        assert len(profile_array.shape) == 2
        assert profile_array.shape[1] == 3

        n = profile_array.shape[0]
        angle = 2 * np.pi / nphi

        nodes_slice = np.concatenate([profile_array, axis.rotate_points(profile_array, angle)])
        faces_slice = np.array([[i, i+n, i+n+1, i+1] for i in range(n-1)])
        body_slice = Mesh(nodes_slice, faces_slice, name=f"slice_of_{name}")
        body_slice.merge_duplicates()
        body_slice.heal_triangles()

        return AxialSymmetricMesh(body_slice, axis=axis, nb_repetitions=nphi - 1, name=name)

    @property
    def first_slice(self):
        return self[0]

    def tree_view(self, fold_symmetry=True, **kwargs):
        if fold_symmetry:
            return (str(self) + '\n' + ' ├─' + self.first_slice.tree_view().replace('\n', '\n │ ') + '\n'
                    + f" └─{len(self)-1} rotated copies of the above {str(self.first_slice)}")
        else:
            return CollectionOfMeshes.tree_view(self, **kwargs)

    def __deepcopy__(self, *args):
        return AxialSymmetricMesh(self.first_slice.copy(), axis=self.axis.copy(), nb_repetitions=len(self) - 1, name=self.name)

    def join_meshes(*meshes, name=None):
        assert all(isinstance(mesh, AxialSymmetricMesh) for mesh in meshes), \
            "Only meshes with the same symmetry can be joined together."
        assert all(meshes[0].axis == mesh.axis for mesh in meshes), \
            "Only axisymmetric meshes with the same symmetry axis can be joined together."
        assert all(len(meshes[0]) == len(mesh) for mesh in meshes), \
            "Only axisymmetric meshes with the same number of elements can be joined together."
        mesh_slice = CollectionOfMeshes([mesh.first_slice for mesh in meshes], name=f"slice_of_{name}" if name is not None else None)
        return AxialSymmetricMesh(mesh_slice, axis=meshes[0].axis, nb_repetitions=len(meshes[0]) - 1, name=name)

    @inplace_transformation
    def translate(self, vector):
        self.axis.translate(vector)
        CollectionOfMeshes.translate(self, vector)
        return self

    @inplace_transformation
    def rotate(self, other_axis: Axis, angle: float):
        self.axis.rotate(other_axis, angle)
        CollectionOfMeshes.rotate(self, other_axis, angle)
        return self

    @inplace_transformation
    def mirror(self, plane: Plane):
        self.axis.mirror(plane)
        CollectionOfMeshes.mirror(self, plane)
        return self




#!/usr/bin/env python
# coding: utf-8
"""Tools for 3D displays with VTK."""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

from typing import Union

from capytaine.meshes.meshes import Mesh
from capytaine.meshes.collections import CollectionOfMeshes
from capytaine.tools.optional_imports import import_optional_dependency

vtk = import_optional_dependency("vtk")

def compute_vtk_polydata(mesh: Union[Mesh, CollectionOfMeshes]):
    """Transform a mesh into vtkPolydata."""

    # Create a vtkPoints object and store the points in it
    points = vtk.vtkPoints()
    for point in mesh.vertices:
        points.InsertNextPoint(point)

    # Create a vtkCellArray to store faces
    faces = vtk.vtkCellArray()
    for face_ids in mesh.faces:
        if face_ids[0] == face_ids[-1]:
            # Triangle
            curface = face_ids[:3]
            vtk_face = vtk.vtkTriangle()
        else:
            # Quadrangle
            curface = face_ids[:4]
            vtk_face = vtk.vtkQuad()

        for idx, id in enumerate(curface):
            vtk_face.GetPointIds().SetId(idx, id)

        faces.InsertNextCell(vtk_face)

    vtk_polydata = vtk.vtkPolyData()
    vtk_polydata.SetPoints(points)
    vtk_polydata.SetPolys(faces)

    return vtk_polydata


def compute_node_data(mesh: Union[Mesh, CollectionOfMeshes],
                      face_data):
    """Transform data defined at the center of the faces to data defined at the nodes of the mesh
    by a simple averaging of the values of the neighboring faces.

    Parameters
    ----------
    mesh: Mesh or CollectionOfMeshes
        the mesh on which the face face_data are defined
    face_data: numpy array of shape (mesh.nb_faces, ...)
        the data defined on the center of the faces of the mesh

    Returns
    -------
    node_data: numpy array of shape (mesh.nb_vertices, ...)
        the same data averaged on the nodes
    """

    import numpy as np

    mesh = mesh.merged()
    assert face_data.shape[0] == mesh.nb_faces

    # Initialize output array
    node_data_shape = (mesh.vertices.shape[0], ) + face_data.shape[1:]
    node_data = np.zeros(node_data_shape, dtype=complex)

    # Keep track of the number of faces near each vertex
    faces_near_nodes_shape = (mesh.vertices.shape[0], ) + (1, ) * len(face_data.shape[1:])
    nb_faces_near_nodes = np.zeros(faces_near_nodes_shape, dtype=np.int8)

    for i, vertices in enumerate(mesh.faces):
        for vertex in vertices:
            nb_faces_near_nodes[vertex] += 1
            node_data[vertex, ...] += face_data[i, ...]

    node_data /= nb_faces_near_nodes
    return node_data




#!/usr/bin/env python
# coding: utf-8
"""Definition of the problems to solve with the BEM solver, and the results of this resolution."""
# Copyright (C) 2017-2020 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging

import numpy as np
import pandas as pd
from scipy.optimize import newton

from capytaine.meshes.collections import CollectionOfMeshes
from capytaine.bem.airy_waves import airy_waves_velocity, froude_krylov_force

LOG = logging.getLogger(__name__)

_default_parameters = {'rho': 1000.0, 'g': 9.81, 'omega': 1.0,
                      'free_surface': 0.0, 'water_depth': np.infty,
                      'convention': "nemoh"}


class LinearPotentialFlowProblem:
    """General class of a potential flow problem.

    Parameters
    ----------
    body: FloatingBody, optional
        The body interacting with the waves
    free_surface: float, optional
        The position of the free surface (accepted values: 0 and np.infty)
    sea_bottom: float, optional
        The position of the sea bottom
    omega: float, optional
        The frequency of the waves in rad/s
    rho: float, optional
        The density of water in kg/m3 (default: 1000.0)
    g: float, optional
        The acceleration of gravity in m/s2 (default: 9.81)
    boundary_condition: np.ndarray of shape (body.mesh.nb_faces,)
        The Neumann boundary condition on the floating body

    TODO: more consistent use of free_surface and sea_bottom vs. water_depth
    """

    def __init__(self, *,
                 body=None,
                 free_surface=_default_parameters['free_surface'],
                 sea_bottom=-_default_parameters['water_depth'],
                 omega=_default_parameters['omega'],
                 rho=_default_parameters['rho'],
                 g=_default_parameters['g'],
                 boundary_condition=None):

        self.body = body
        self.free_surface = free_surface
        self.sea_bottom = sea_bottom
        self.omega = omega
        self.rho = rho
        self.g = g
        self.boundary_condition = boundary_condition

        self._check_data()

    def _check_data(self):
        """Sanity checks on the data."""

        if self.free_surface not in [0.0, np.infty]:
            raise NotImplementedError(
                f"Free surface is {self.free_surface}. "
                "Only z=0 and z=∞ are accepted values for the free surface position."
            )

        if self.free_surface == np.infty and self.sea_bottom != -np.infty:
            raise NotImplementedError(
                "Problems with a sea bottom but no free surface have not been implemented."
            )

        if self.free_surface < self.sea_bottom:
            raise ValueError("Sea bottom is above the free surface.")

        if self.omega in {0, np.infty} and self.depth != np.infty:
            raise NotImplementedError(
                f"omega={self.omega} is only implemented for infinite depth."
            )

        if self.body is not None:
            if ((isinstance(self.body.mesh, CollectionOfMeshes) and len(self.body.mesh) == 0)
                    or len(self.body.mesh.faces) == 0):
                raise ValueError(f"The mesh of the body {self.body.name} is empty.")

            if (any(self.body.mesh.faces_centers[:, 2] >= self.free_surface)
                    or any(self.body.mesh.faces_centers[:, 2] <= self.sea_bottom)):

                LOG.warning(
                    f"The mesh of the body {self.body.name} is not inside the domain.\n"
                    "Check the position of the free_surface and the sea_bottom\n"
                    "or use body.keep_immersed_part() to clip the mesh."
                )

            if self.wavelength < 8*self.body.mesh.faces_radiuses.max():
                LOG.warning(f"Mesh resolution for {self}:\n"
                        f"The resolution of the mesh '{self.body.mesh.name}' of the body '{self.body.name}' "
                        f"might be insufficient for the wavelength λ={self.wavelength:.2e}.\n"
                        f"This warning appears because the largest panel of this mesh has radius {self.body.mesh.faces_radiuses.max():.2e} > λ/8."
                        )

        if self.boundary_condition is not None:
            if len(self.boundary_condition.shape) != 1:
                raise ValueError("Expected a 1-dimensional array as boundary_condition")

            if self.boundary_condition.shape[0] != self.body.mesh.nb_faces:
                raise ValueError(
                    f"The shape of the boundary condition ({self.boundary_condition.shape})"
                    f"does not match the number of faces of the mesh ({self.body.mesh.nb_faces})."
                )

    @property
    def body_name(self):
        return self.body.name if self.body is not None else 'None'

    def _asdict(self):
        return {"body_name": self.body_name,
                "water_depth": self.water_depth,
                "omega": self.omega,
                "rho": self.rho,
                "g": self.g}

    @staticmethod
    def _group_for_parallel_resolution(problems):
        """Given a list of problems, returns a list of groups of problems, such
        that each group should be executed in the same process to benefit from
        caching.
        """
        problems_params = pd.DataFrame([pb._asdict() for pb in problems])
        groups_of_indices = problems_params.groupby(["body_name", "water_depth", "omega", "rho", "g"]).groups.values()
        groups_of_problems = [[problems[i] for i in grp] for grp in groups_of_indices]
        return groups_of_problems

    def __str__(self):
        """Do not display default values in str(problem)."""
        parameters = [f"body={self.body_name}",
                      f"omega={self.omega:.3f}",
                      f"depth={self.depth}"]
        try:
            parameters.extend(self._str_other_attributes())
        except AttributeError:
            pass

        if not self.free_surface == _default_parameters['free_surface']:
            parameters.append(f"free_surface={self.free_surface}")
        if not self.g == _default_parameters['g']:
            parameters.append(f"g={self.g}")
        if not self.rho == _default_parameters['rho']:
            parameters.append(f"rho={self.rho}")

        return self.__class__.__name__ + "(" + ', '.join(parameters) + ")"

    def _astuple(self):
        return (self.body, self.free_surface, self.sea_bottom, self.omega, self.rho, self.g)

    def __eq__(self, other):
        if isinstance(other, LinearPotentialFlowProblem):
            return self._astuple() == other._astuple()
        else:
            return NotImplemented

    def __lt__(self, other):
        # Arbitrary order. Used for ordering of problems: problems with same body are grouped together.
        if isinstance(other, LinearPotentialFlowProblem):
            return self._astuple()[:6] < other._astuple()[:6]
            # Not the whole tuple, because when using inheriting classes,
            # "radiating_dof" cannot be compared with "wave_direction"
        else:
            return NotImplemented

    @property
    def water_depth(self):
        return self.free_surface - self.sea_bottom

    @property
    def depth(self):
        return self.water_depth

    @property
    def wavenumber(self):
        if self.depth == np.infty or self.omega**2*self.depth/self.g > 20:
            return self.omega**2/self.g
        else:
            return newton(lambda x: x*np.tanh(x) - self.omega**2*self.depth/self.g, x0=1.0)/self.depth

    @property
    def wavelength(self):
        if self.wavenumber == 0.0:
            return np.infty
        else:
            return 2*np.pi/self.wavenumber

    @property
    def period(self):
        if self.omega == 0.0:
            return np.infty
        else:
            return 2*np.pi/self.omega

    @property
    def dimensionless_omega(self):
        if self.depth != np.infty:
            return self.omega**2*self.depth/self.g
        else:
            raise AttributeError("Dimensionless omega is defined only for finite depth problems.")

    @property
    def dimensionless_wavenumber(self):
        if self.depth != np.infty:
            return self.wavenumber*self.depth
        else:
            raise AttributeError("Dimensionless wavenumber is defined only for finite depth problems.")

    @property
    def influenced_dofs(self):
        # TODO: let the user choose the influenced dofs
        return self.body.dofs if self.body is not None else set()

    def make_results_container(self):
        return LinearPotentialFlowResult(self)


class DiffractionProblem(LinearPotentialFlowProblem):
    """Particular LinearPotentialFlowProblem with boundary conditions
    computed from an incoming Airy wave."""

    def __init__(self, *,
                 body=None,
                 free_surface=_default_parameters['free_surface'],
                 sea_bottom=-_default_parameters['water_depth'],
                 omega=_default_parameters['omega'],
                 rho=_default_parameters['rho'],
                 g=_default_parameters['g'],
                 wave_direction=0.0,
                 convention=_default_parameters['convention']):

        self.wave_direction = wave_direction
        self.convention = convention

        super().__init__(body=body, free_surface=free_surface, sea_bottom=sea_bottom,
                         omega=omega, rho=rho, g=g)

        if not (-2*np.pi-1e-3 <= self.wave_direction <= 2*np.pi+1e-3):
            LOG.warning(f"The value {self.wave_direction} has been provided for the wave direction, and it does not look like an angle in radians. "
                         "The wave direction in Capytaine is defined in radians and not in degrees, so the result might not be what you expect. "
                         "If you were actually giving an angle in radians, use the modulo operator to give a value between -2π and 2π to disable this warning.")

        if self.body is not None:

            self.boundary_condition = -(
                    airy_waves_velocity(self.body.mesh.faces_centers, self, convention=self.convention)
                    * self.body.mesh.faces_normals
            ).sum(axis=1)

            if len(self.body.dofs) == 0:
                LOG.warning(f"The body {self.body.name} used in diffraction problem has no dofs!")

    def _astuple(self):
        return super()._astuple() + (self.wave_direction,)

    def _asdict(self):
        d = super()._asdict()
        d["wave_direction"] = self.wave_direction
        d["convention"] = self.convention
        return d

    def _str_other_attributes(self):
        return [f"wave_direction={self.wave_direction:.3f}"]

    def make_results_container(self):
        return DiffractionResult(self)


class RadiationProblem(LinearPotentialFlowProblem):
    """Particular LinearPotentialFlowProblem whose boundary conditions have
    been computed from the degree of freedom of the body."""

    def __init__(self, *, body=None,
                 free_surface=_default_parameters['free_surface'],
                 sea_bottom=-_default_parameters['water_depth'],
                 omega=_default_parameters['omega'],
                 rho=_default_parameters['rho'],
                 g=_default_parameters['g'],
                 radiating_dof=None):

        self.radiating_dof = radiating_dof

        super().__init__(body=body, free_surface=free_surface, sea_bottom=sea_bottom,
                         omega=omega, rho=rho, g=g)

        if self.body is not None:

            if len(self.body.dofs) == 0:
                raise ValueError(f"Body {self.body.name} does not have any degrees of freedom.")

            if self.radiating_dof is None:
                self.radiating_dof = next(iter(self.body.dofs))

            if self.radiating_dof not in self.body.dofs:
                LOG.error(f"In {self}: the radiating degree of freedom {self.radiating_dof} is not one of"
                          f"the degrees of freedom of the body.\n"
                          f"The dofs of the body are {list(self.body.dofs.keys())}")
                raise ValueError("Unrecognized degree of freedom name.")

            dof = self.body.dofs[self.radiating_dof]
            self.boundary_condition = np.sum(dof * self.body.mesh.faces_normals, axis=1)

    def _astuple(self):
        return super()._astuple() + (self.radiating_dof,)

    def _asdict(self):
        d = super()._asdict()
        d["radiating_dof"] = self.radiating_dof
        return d

    def _str_other_attributes(self):
        return [f"radiating_dof={self.radiating_dof}"]

    def make_results_container(self):
        return RadiationResult(self)


class LinearPotentialFlowResult:

    def __init__(self, problem):
        self.problem = problem

        self.sources = None
        self.potential = None
        self.fs_elevation = {}

        # Copy data from problem
        self.body               = self.problem.body
        self.free_surface       = self.problem.free_surface
        self.sea_bottom         = self.problem.sea_bottom
        self.omega              = self.problem.omega
        self.rho                = self.problem.rho
        self.g                  = self.problem.g
        self.boundary_condition = self.problem.boundary_condition
        self.water_depth        = self.problem.water_depth
        self.depth              = self.problem.depth
        self.wavenumber         = self.problem.wavenumber
        self.wavelength         = self.problem.wavelength
        self.period             = self.problem.period
        self.body_name          = self.problem.body_name
        self.influenced_dofs    = self.problem.influenced_dofs

    __str__ = LinearPotentialFlowProblem.__str__


class DiffractionResult(LinearPotentialFlowResult):

    def __init__(self, problem):
        super().__init__(problem)
        self.wave_direction = self.problem.wave_direction
        self.forces = {}

    def store_force(self, dof, force):
        self.forces[dof] = 1j*self.omega*force

    @property
    def records(self):
        params = self.problem._asdict()
        FK = froude_krylov_force(self.problem)
        return [dict(**params,
                     influenced_dof=dof,
                     diffraction_force=self.forces[dof],
                     Froude_Krylov_force=FK[dof])
                for dof in self.influenced_dofs]


class RadiationResult(LinearPotentialFlowResult):

    def __init__(self, problem):
        super().__init__(problem)
        self.radiating_dof = self.problem.radiating_dof
        self.added_masses = {}
        self.radiation_dampings = {}

    def store_force(self, dof, force):
        self.added_masses[dof] = force.real
        if self.problem.omega == np.infty:
            self.radiation_dampings[dof] = 0
        else:
            self.radiation_dampings[dof] = self.problem.omega * force.imag

    @property
    def records(self):
        params = self.problem._asdict()
        return [dict(params,
                     influenced_dof=dof,
                     added_mass=self.added_masses[dof],
                     radiation_damping=self.radiation_dampings[dof])
                for dof in self.influenced_dofs]



#!/usr/bin/env python
# coding: utf-8
"""Tools for surface integrals.
Based on meshmagick <https://github.com/LHEEA/meshmagick> by François Rongère.
"""
# Copyright (C) 2017-2019 Matthieu Ancellin, based on the work of François Rongère
# See LICENSE file at <https://github.com/mancellin/capytaine>

import numpy as np


def compute_faces_integrals(mesh):
    """Compute intergrals on the faces."""

    surface_integrals = np.zeros((15, mesh.nb_faces), dtype=float)

    # First triangles
    if mesh.nb_triangles > 0:
        triangles_ids = mesh.triangles_ids
        triangles_vertices = mesh.vertices[mesh._faces[triangles_ids][:, :3]]
        surface_integrals[:, triangles_ids] = _compute_triangles_integrals(triangles_vertices)

    # Now quadrangles by splitting them up
    if mesh.nb_quadrangles > 0:
        quadrangles_ids = mesh.quadrangles_ids
        quadrangles = mesh.faces[quadrangles_ids]

        # First pass
        surface_integrals[:, quadrangles_ids] = \
            _compute_triangles_integrals(mesh._vertices[quadrangles[:, (0, 1, 2)]])

        # Second pass
        surface_integrals[:, quadrangles_ids] += \
            _compute_triangles_integrals(mesh._vertices[quadrangles[:, (0, 2, 3)]])

    return surface_integrals


def _compute_triangles_integrals(triangles_vertices):
    """Performs the computation of the various interesting surface integrals.

    Notes
    -----
    triangles vertices must describe by increasing dimension from the general to the particular:
    dimension 0 : information about each facet -- triangles_vertices[0] -> facet 0)
    dimension 1 : information about each facet vertex -- triangles_vertices[0, 1] -> vertex 1 of facet 0
    dimension 2 : information on each vertex coordinate -- triangles_vertices[0, 1, 2] -> z coordinate of vertex 1 of facet 0

    Todo
    ----
    Explicit the integrals
    """

    nb_triangles = triangles_vertices.shape[0]
    s_int = np.zeros((15, nb_triangles), dtype=float)

    point_0, point_1, point_2 = list(map(_3DPointsArray, np.rollaxis(triangles_vertices, 1, 0)))

    t0 = point_0 + point_1                    # t0 = p0 + p1
    f1 = t0 + point_2                         # f1 = p0 + p1 + p2
    t1 = point_0 * point_0                    # t1 = p0**2
    t2 = t1 + point_1*t0                      # t2 = p0**2 + p1**2 + p1p0
    f2 = t2 + point_2*f1                      # f2 = p0**2 + p1**2 + p2**2 + p0*p1 + p0*p2 + p1*p2
    f3 = point_0*t1 + point_1*t2 + point_2*f2 # f3 = p0**3 + p1**3 + p2**3 + p1*p0**2 + ...
    g0 = f2 + point_0 * (f1 + point_0)        # g0 = 3p0**2 + p1**2  + p2**2  + 2p0*p1 + 2p0*p2 + p1*p2
    g1 = f2 + point_1 * (f1 + point_1)        # g1 = p0**2  + 3p1**2 + p2**2  + 2p0*p1 + p0*p2  + 2p1*p2
    g2 = f2 + point_2 * (f1 + point_2)        # g2 = p0**2  + p1**2  + 3p2**2 + p0*p1  + 2p0*p2 + 2p1*p2

    e1 = point_1 - point_0
    e2 = point_2 - point_0

    delta = np.linalg.norm(np.cross(e1, e2), axis=1)

    s_int[0:3] = np.einsum('i, ij -> ji', delta, f1) / 6.

    s_int[3] = delta * (6.*point_0.y*point_0.z + 3*(point_1.y*point_1.z + point_2.y*point_2.z) - point_0.y*f1[:, 2] - point_0.z*f1[:, 1]) / 12.
    s_int[4] = delta * (6.*point_0.x*point_0.z + 3*(point_1.x*point_1.z + point_2.x*point_2.z) - point_0.x*f1[:, 2] - point_0.z*f1[:, 0]) / 12.
    s_int[5] = delta * (6.*point_0.x*point_0.y + 3*(point_1.x*point_1.y + point_2.x*point_2.y) - point_0.x*f1[:, 1] - point_0.y*f1[:, 0]) / 12.

    s_int[6:9] = np.einsum('i, ij -> ji', delta, f2) / 12.
    s_int[9:12] = np.einsum('i, ij -> ji', delta, f3) / 20.

    s_int[12] = delta * (point_0.y*g0[:, 0] + point_1.y*g1[:, 0] + point_2.y*g2[:, 0]) / 60.
    s_int[13] = delta * (point_0.z*g0[:, 1] + point_1.z*g1[:, 1] + point_2.z*g2[:, 1]) / 60.
    s_int[14] = delta * (point_0.x*g0[:, 2] + point_1.x*g1[:, 2] + point_2.x*g2[:, 2]) / 60.

    return s_int


class _3DPointsArray(np.ndarray):
    def __new__(cls, points):
        obj = np.asarray(points).view(cls)
        cls.x = property(fget=lambda cls: cls[:, 0])
        cls.y = property(fget=lambda cls: cls[:, 1])
        cls.z = property(fget=lambda cls: cls[:, 2])
        return obj



#!/usr/bin/env python
# coding: utf-8
"""Abstract structure of a class used to compute the Green function"""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

from abc import ABC, abstractmethod

class AbstractGreenFunction(ABC):
    """Abstract method to evaluate the Green function."""

    @abstractmethod
    def evaluate(self, mesh1, mesh2, free_surface, sea_bottom, wavenumber):
        pass




#!/usr/bin/env python
# coding: utf-8

import os

from setuptools import dist
dist.Distribution().fetch_build_eggs(['numpy'])

from numpy.distutils.core import Extension, setup

########################
#  Fortran extensions  #
########################

Delhommeau_source = [
        "capytaine/green_functions/libDelhommeau/src/constants.f90",
        "capytaine/green_functions/libDelhommeau/src/Delhommeau_integrals.f90",
        "capytaine/green_functions/libDelhommeau/src/old_Prony_decomposition.f90",
        "capytaine/green_functions/libDelhommeau/src/Green_Rankine.f90",
        "capytaine/green_functions/libDelhommeau/src/Green_wave.f90",
        "capytaine/green_functions/libDelhommeau/src/matrices.f90",
    ]

Delhommeau_extension = Extension(
    name="capytaine.green_functions.Delhommeau_f90",
    sources=Delhommeau_source,
    extra_f90_compile_args=['-O2', '-fopenmp', '-cpp'],
    extra_link_args=['-fopenmp'],
    # # Uncomment the following lines to get more verbose output from f2py.
    # define_macros=[
    #     ('F2PY_REPORT_ATEXIT', 1),
    #     ('F2PY_REPORT_ON_ARRAY_COPY', 1),
    # ],
)

XieDelhommeau_extension = Extension(
    name="capytaine.green_functions.XieDelhommeau_f90",
    sources=Delhommeau_source,
    extra_f90_compile_args=['-O2', '-fopenmp', '-cpp', '-DXIE_CORRECTION'],
    extra_link_args=['-fopenmp'],
    # # Uncomment the following lines to get more verbose output from f2py.
    # define_macros=[
    #     ('F2PY_REPORT_ATEXIT', 1),
    #     ('F2PY_REPORT_ON_ARRAY_COPY', 1),
    # ],
)


########################################################
#  Read version number and other info in __about__.py  #
########################################################

base_dir = os.path.dirname(__file__)
src_dir = os.path.join(base_dir, "capytaine")

about = {}
with open(os.path.join(src_dir, "__about__.py")) as f:
    exec(f.read(), about)


##########
#  Main  #
##########

if __name__ == "__main__":
    setup(name=about["__title__"],
          version=about["__version__"],
          description=about["__description__"],
          author=about["__author__"],
          license=about["__license__"],
          url=about["__uri__"],
          packages=[
              'capytaine',
              'capytaine.meshes',
              'capytaine.matrices',
              'capytaine.bodies',
              'capytaine.bodies.predefined',
              'capytaine.bem',
              'capytaine.green_functions',
              'capytaine.post_pro',
              'capytaine.ui',
              'capytaine.ui.vtk',
              'capytaine.io',
              'capytaine.tools',
          ],
          install_requires=[
              'numpy',
              'scipy',
              'pandas>=1.3',
              'xarray',
          ],
          extras_require={
            'develop': [
              'pytest',
              'hypothesis',
              'ipython',
              'matplotlib',
              'vtk',
              'meshio',
              'pygmsh',
              'gmsh',
              'quadpy',
              'bemio @ git+https://github.com/michaelcdevin/bemio.git@master-python3#egg=bemio',
              'sphinx',
              'sphinxcontrib-proof',
            ],
            'ci': [
              'pytest',
              'hypothesis',
                ],
            'extra': [
              'ipython',
              'matplotlib',
              'vtk',
              'meshio',
              'pygmsh',
              'gmsh',
              'quadpy',
              'bemio @ git+https://github.com/michaelcdevin/bemio.git@master-python3#egg=bemio',
            ]
          },
          entry_points={
              'console_scripts': [
                  'capytaine=capytaine.ui.cli:main',
              ],
          },
          ext_modules=[
              Delhommeau_extension,
              XieDelhommeau_extension,
          ],
          )
